import admm
import pock_chambolle as pc
import half_quadratic_splitting as hqs
import linearized_admm as ladmm

from .absorb import absorb_lin_op, absorb_offset
from .problem import Problem
from .equil import equil
from .merge import can_merge, merge_fns



from .lin_op import LinOp
import numpy as np
from proximal.utils.utils import *

class mul_color(LinOp):
    """Color transform as blockwise 3x3 transform

       mode can be "opp" (opponenent color space)
       "yuv" (yuv color space)
       3 x 3 array or matrix as general transform
    """
    def __init__(self, arg, mode):
        #General transform or predefined mode
        if isinstance(mode, np.ndarray) or isinstance(mode, np.matrix):

            #General transform
            self.TC = np.asarray(mode)

            if self.TC.shape != (3,3):
                print >> sys.stderr, "Error, color matrix is not 3x3."
                sys.exit(1)

        else:

            #Check for predefined transforms
            if mode == 'opp':
                self.TC = np.array([[1./3., 1./3., 1./3.], [0.5, 0.0, -0.5], [0.25, -0.5, 0.25]], dtype=np.float32, order='F')
            elif mode == 'yuv':
                self.TC = np.array([[0.299, 0.587, 0.114], [-0.16873660714285, -0.33126339285715, 0.5],[0.5 ,-0.4186875, -0.0813125]], dtype=np.float32, order='F')
            else:
                print >> sys.stderr, "Error, unsupported color mode."
                sys.exit(1)

        #Check for the shape
        if len(arg.shape) != 3 or arg.shape[2] != 3:
            print >> sys.stderr, "Error, color transform needs a 3 channel image."
            sys.exit(1)

        super(mul_color, self).__init__([arg], arg.shape)

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """

        inimg = inputs[0]
        inimg_reshaped = inimg.reshape((inimg.shape[0] * inimg.shape[1], inimg.shape[2]))
        result = np.dot(self.TC, inimg_reshaped.T ).T.reshape(inimg.shape)
        np.copyto( outputs[0], result )

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """

        inimg = inputs[0]
        inimg_reshaped = inimg.reshape((inimg.shape[0] * inimg.shape[1], inimg.shape[2]))
        result = np.dot( self.TC.T, inimg_reshaped.T).T.reshape(inimg.shape)
        np.copyto( outputs[0], result )

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return input_mags[0]*np.linalg.norm(self.TC, 2)



from .prox_fn import ProxFn
import numpy as np

class nonneg(ProxFn):
    """The function I(x >= 0).
    """
    def _prox(self, rho, v, *args, **kwargs):
        """x = pos(x).
        """
        np.maximum(v, 0, v)
        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        if np.min(v) > -1e-3:
            return 0
        else:
            return np.inf

class weighted_nonneg(nonneg):
    """The function I(W*x >= 0).
    """
    def __init__(self, lin_op, weight, **kwargs):
        self.weight = weight
        super(weighted_nonneg, self).__init__(lin_op, **kwargs)

    def _prox(self, rho, v, *args, **kwargs):
        """x = pos(x).
        """
        idxs = self.weight != 0.
        v[idxs] = np.maximum(self.weight[idxs]*v[idxs], 0.)/self.weight[idxs]
        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return super(weighted_nonneg, self)._eval(self.weight*v)

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.weight]

class masked_nonneg(nonneg):
    """The function I(x >= 0) for M == 1 else I(x == 0)
       Converts any non-binary mask to a binary one.
    """
    def __init__(self, lin_op, mask, **kwargs):
        self.mask = mask
        self.mask[ self.mask < 0.5 ] = 0.0
        self.mask[ self.mask >= 0.5 ] = 1.0
        super(masked_nonneg, self).__init__(lin_op, **kwargs)

    def _prox(self, rho, v, *args, **kwargs):
        """x = pos(x).
        """
        np.maximum(v, 0, v)
        v *= self.mask
        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        if np.min( self.mask* v) > -1e-3 or np.max(np.absolute( (1.0 - self.mask)* v)) > -1e-3:
            return 0
        else:
            return np.inf

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.mask]



from .prox_fn import ProxFn
import numpy as np
from proximal.utils.utils import *
from proximal.halide.halide import *
import cv2

class patch_NLM(ProxFn):
    """The function for NLM patch prior
    """
    def __init__(self, lin_op, params=None, **kwargs):

        self.params = params
        if params is None:
            self.sigma_fixed = 0.6
            self.sigma_scale = 1.5  * 1 * 4
            self.templateWindowSizeNLM = 1 * 2 + 1
            self.searchWindowSizeNLM = 5 * 2 + 1
            self.gamma_trans= 1.0
            self.prior = 1
        else:
            self.sigma_fixed = params[0]
            self.sigma_scale = params[1]
            self.templateWindowSizeNLM = params[2]
            self.searchWindowSizeNLM = params[3]
            self.gamma_trans = params[4]
            self.prior = params[5]

        #Halide
        if len(lin_op.shape) == 3 and lin_op.shape[2] == 3:
            self.tmpout = np.zeros(lin_op.shape, dtype=np.float32, order='F')
            self.paramsh = np.asfortranarray( np.array([self.sigma_fixed, 1.0, self.sigma_scale, self.prior],
                                            dtype=np.float32)[..., np.newaxis] )

        super(patch_NLM, self).__init__(lin_op, **kwargs)

    def _prox(self, rho, v, *args, **kwargs):
        """x = denoise_gaussian_NLM( tonemap(v), sqrt(1/rho))
        """

        if self.implementation == Impl['halide'] and len(self.lin_op.shape) == 3 and self.lin_op.shape[2] == 3:

            #Halide implementation
            tmpin = np.asfortranarray(v.astype(np.float32))
            Halide('prox_NLM.cpp').prox_NLM(tmpin, 1./rho, self.paramsh, self.tmpout)
            np.copyto(v, self.tmpout)

        else:

            #Compute sigma
            sigma = np.sqrt(1.0/rho)

            #Fixed sigma if wanted
            sigma_estim = sigma
            if self.sigma_fixed > 0.0:
                sigma_estim = self.sigma_fixed / 30.0 * self.sigma_scale;

            #Params
            print "Prox NLM params are: [sigma ={0} prior={1} sigma_scale={2}]".format(sigma_estim, self.prior, self.sigma_scale)

            #Scale d
            v = v.copy()
            v_min = np.amin(v)
            v_max = np.amax(v)
            v_max = np.maximum(v_max, v_min + 0.01)

            #Scale and offset parameters d
            v -= v_min
            v /= (v_max - v_min)

            #Denoising params
            sigma_luma = sigma_estim
            sigma_color = sigma_estim
            if self.prior > 0.5 :
                sigma_luma =  sigma_estim * 1.3 #NLM color stronger on luma
                sigma_color = sigma_color * 3.0

            #Transform
            if self.gamma_trans != 1.0:
                vuint = np.uint8(np.clip( v**self.gamma_trans * 255.0,0, 255)) #Quadratic tranform to account for gamma
            else:
                vuint = np.uint8(np.clip( v * 255.0,0.0, 255.0)) #Quadratic tranform to account for gamma

            if self.prior <= 0.5 :

                vdstuint = cv2.fastNlMeansDenoising( vuint, None,
                                                        sigma_luma * 255.0,
                                                        self.templateWindowSizeNLM, self.searchWindowSizeNLM)
            else:

                vdstuint = cv2.fastNlMeansDenoisingColored( vuint, None,
                                                        sigma_luma * 255.0, sigma_color * 255.,
                                                        self.templateWindowSizeNLM, self.searchWindowSizeNLM)

            #Convert to float and inverse scale
            if self.gamma_trans != 1.0:
                dst = ((vdstuint.astype(np.float32) / 255.0)**(1.0/self.gamma_trans)) * (v_max - v_min) + v_min
            else:
                dst = vdstuint.astype(np.float32) / 255.0 * (v_max - v_min) + v_min
            np.copyto( v, dst)

        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return np.inf #TODO: IGNORE FOR NOW

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.params]



# Absorb linear operators into proximal operators.

from proximal.lin_ops import *
from proximal.prox_fns import *
import numpy as np
import copy as cp

WEIGHTED = {nonneg: weighted_nonneg,
            norm1: weighted_norm1,
            sum_squares: weighted_sum_squares,
            poisson_norm: weighted_poisson_norm,
            group_norm1: weighted_group_norm1}


def absorb_all_lin_ops(prox_funcs):
    """Repeatedy absorb lin ops.
    """
    new_proxes = []
    ready = prox_funcs[:]
    while len(ready) > 0:
        curr = ready.pop(0)
        absorbed = absorb_lin_op(curr)
        if len(absorbed) == 1 and absorbed[0] == curr:
            new_proxes.append(absorbed[0])
        else:
            ready += absorbed
    return new_proxes


def absorb_lin_op(prox_fn):
    """If possible moves the top level lin op argument
       into the prox operator.

       For example, elementwise multiplication can be folded into
       a separable function's prox.
    """
    # Never eliminate variables.
    if isinstance(prox_fn.lin_op, Variable):
        return [prox_fn]
    # Absorb a lin op into sum_entries/zero.
    if type(prox_fn) == zero_prox and prox_fn.gamma == 0:
        outputs = []
        inputs = [prox_fn.c]
        for arg in prox_fn.lin_op.input_nodes:
            outputs.append(np.zeros(arg.shape))
        prox_fn.lin_op.adjoint(inputs, outputs)
        new_proxes = []
        for output, arg in zip(outputs, prox_fn.lin_op.input_nodes):
            new_proxes.append( prox_fn.copy(arg, c=output) )
        return new_proxes
    # Fold scaling into the function.
    if isinstance(prox_fn.lin_op, mul_elemwise):
        op_weight = prox_fn.lin_op.weight
        def get_new_prox(prox_type, args):
            new_prox = prox_type(*args)
            copy_prox_fn(new_prox, prox_fn)
            idxs = op_weight != 0
            new_prox.b[idxs] = prox_fn.b[idxs]/op_weight[idxs]
            new_prox.c = prox_fn.c*op_weight
            return [new_prox]

        if type(prox_fn) in WEIGHTED.keys() and prox_fn.gamma == 0:
            args = [prox_fn.lin_op.input_nodes[0]] + prox_fn.get_data() + \
                   [op_weight]
            return get_new_prox(WEIGHTED[type(prox_fn)], args)
        elif type(prox_fn) in WEIGHTED.values() and prox_fn.gamma == 0:
            args = [prox_fn.lin_op.input_nodes[0]] + prox_fn.get_data()
            args[-1] = args[-1]*op_weight
            return get_new_prox(type(prox_fn), args)
    # Fold scalar into the function.
    if isinstance(prox_fn.lin_op, scale):
        scalar = prox_fn.lin_op.scalar
        new_prox = prox_fn.copy(prox_fn.lin_op.input_nodes[0],
            beta=prox_fn.beta*scalar, b=prox_fn.b/scalar, c=prox_fn.c*scalar,
            gamma=prox_fn.gamma*scalar**2)
        return [new_prox]
    # No change.
    return [prox_fn]


def copy_prox_fn(dst_prox, src_prox):
    """Copy the optional parameters from src_prox to dst_prox.
    """
    dst_prox.alpha = src_prox.alpha
    dst_prox.beta = src_prox.beta
    dst_prox.gamma = src_prox.gamma
    dst_prox.b = src_prox.b
    dst_prox.c = src_prox.c
    dst_prox.d = src_prox.d

def copy_non_var(lin_op):
    """If not a variable, returns a shallow copy.
    """
    if isinstance(lin_op, Variable):
        return lin_op
    else:
        return cp.copy(lin_op)

def absorb_offset(prox_fn):
    """Absorb the constant offset into the b term and zero out constants in lin op.
    """
    # Short circuit if no constant leaves.
    if len(prox_fn.lin_op.constants()) == 0:
        return prox_fn
    new_b = -prox_fn.lin_op.get_offset()
    # Zero out constants.
    new_lin_op = copy_non_var(prox_fn.lin_op)
    ready = [new_lin_op]
    while len(ready) > 0:
        curr = ready.pop(0)
        for idx, arg in enumerate(curr.input_nodes):
            if isinstance(arg, Constant):
                curr.input_nodes[idx] = Constant(np.zeros(arg.shape))
            # Don't copy variables.
            else:
                curr.input_nodes[idx] = copy_non_var(arg)
            ready.append(curr.input_nodes[idx])
    return prox_fn.copy(new_lin_op, b=new_b + prox_fn.b)



# Merge proximal operators together.

from proximal.lin_ops import *
from proximal.prox_fns import *
import numpy as np

def merge_all(prox_fns):
    """Merge as many prox functions as possible.
    """
    while True:
        merged = []
        new_prox_fns = []
        no_merges = True
        for i in range(len(prox_fns)):
            for j in range(i+1, len(prox_fns)):
                if prox_fns[i] not in merged and prox_fns[j] not in merged and \
                   can_merge(prox_fns[i], prox_fns[j]):
                    no_merges = False
                    merged += [prox_fns[i], prox_fns[j]]
                    new_prox_fns.append( merge_fns(prox_fns[i], prox_fns[j]) )
        if no_merges:
            break
        prox_fns = new_prox_fns + [fn for fn in prox_fns if fn not in merged]

    return prox_fns

def can_merge(lh_prox, rh_prox):
    """Can lh_prox and rh_prox be merged into a single function?
    """
    # Lin ops must be the same.
    if lh_prox.lin_op == rh_prox.lin_op:
        if type(lh_prox) == zero_prox or type(rh_prox) == zero_prox:
            return True
        elif type(lh_prox) == sum_squares or type(rh_prox) == sum_squares:
            return True

    return False

def merge_fns(lh_prox, rh_prox):
    """Merge the two functions into a single function.
    """
    assert can_merge(lh_prox, rh_prox)
    new_c = lh_prox.c + rh_prox.c
    new_gamma = lh_prox.gamma + rh_prox.gamma
    new_d = lh_prox.d + rh_prox.d
    args = [lh_prox, rh_prox]
    arg_types = [type(lh_prox), type(rh_prox)]
    # Merge a linear term into the other proxable function.
    if zero_prox in arg_types:
        to_copy = args[1-arg_types.index(zero_prox)]
        return to_copy.copy(c=new_c, gamma=new_gamma, d=new_d)
    # Merge a sum squares term into the other proxable function.
    elif sum_squares in arg_types:
        idx = arg_types.index(sum_squares)
        sq_fn = args[idx]
        to_copy = args[1-idx]
        coeff = sq_fn.alpha*sq_fn.beta
        return to_copy.copy(c=new_c - 2*coeff*sq_fn.b,
            gamma=new_gamma + coeff*sq_fn.beta,
            d=new_d + np.square(sq_fn.b).sum())
    else:
        raise ValueError("Unknown merge strategy.")



from .lin_op import LinOp
import numpy as np
from proximal.utils.utils import *
from proximal.halide.halide import *

class conv(LinOp):
    """Circular convolution of the input with a kernel.
    """
    def __init__(self, kernel, arg, dims = None, implem = None):
        self.kernel = kernel
        self.dims = dims if dims < len(arg.shape) else None
        self.forward_kernel = psf2otf(kernel, arg.shape, dims)
        self.adjoint_kernel = self.forward_kernel.conj()
        self.initialized = False
        #Set implementation in super-class
        super(conv, self).__init__([arg], arg.shape, implem)

    # TODO hack. Either add to every LinOp or make copy function like for ProxFns.
    def init_kernel(self):
        if not self.initialized:
            arg = self.input_nodes[0]
            #Replicate kernel for multichannel deconv
            if len(arg.shape) == 3 and len(self.kernel.shape) == 2:
                self.kernel = np.stack( (self.kernel,) * arg.shape[2], axis=-1 )

            #Halide kernel
            if self.implementation == Impl['halide'] and \
               ( len(arg.shape) == 2 or (len(arg.shape) == 3 and arg.dims == 2) ):
                self.kernel = np.asfortranarray(self.kernel.astype(np.float32))
                #Halide FFT (pack into diag)
                #TODO: FIX IMREAL LATER
                hsize = arg.shape if len(arg.shape) == 3 else arg.shape + (1,)
                output_fft_tmp = np.zeros( ((hsize[0] + 1)/2 + 1, hsize[1],hsize[2], 2),
                                           dtype=np.float32, order='F' );
                Halide('fft2_r2c.cpp').fft2_r2c(self.kernel, self.kernel.shape[1]/2,
                                                self.kernel.shape[0]/2, output_fft_tmp)
                self.forward_kernel[:] = 0.

                if len(arg.shape) == 2:
                    self.forward_kernel[0:(hsize[0] + 1)/2 + 1, ...] = 1j* output_fft_tmp[...,0, 1]
                    self.forward_kernel[0:(hsize[0] + 1)/2 + 1, ...] += output_fft_tmp[...,0, 0]
                else:
                    self.forward_kernel[0:(hsize[0] + 1)/2 + 1, ...] = 1j* output_fft_tmp[..., 1]
                    self.forward_kernel[0:(hsize[0] + 1)/2 + 1, ...] += output_fft_tmp[..., 0]

            self.tmpout = np.zeros(arg.shape, dtype=np.float32, order='F')
            self.initialized = True

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        self.init_kernel()
        if self.implementation == Impl['halide'] and \
            ( len(self.shape) == 2 or (len(self.shape) == 3 and self.dims == 2) ):

            #Halide implementation
            tmpin = np.asfortranarray( inputs[0].astype(np.float32) )
            Halide('A_conv.cpp').A_conv( tmpin, self.kernel, self.tmpout ) #Call
            np.copyto(outputs[0], self.tmpout)

        else:

            #Default numpy using FFT
            X = fftd(inputs[0], self.dims)
            X *= self.forward_kernel
            np.copyto(outputs[0], ifftd(X, self.dims).real)

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        self.init_kernel()
        if self.implementation == Impl['halide'] and \
            ( len(self.shape) == 2 or (len(self.shape) == 3 and self.dims == 2) ):

            #Halide implementation
            tmpin = np.asfortranarray( inputs[0].astype(np.float32) )
            Halide('At_conv.cpp').At_conv( tmpin, self.kernel, self.tmpout ) #Call
            np.copyto(outputs[0], self.tmpout)

        else:

            #Default numpy using FFT
            U = fftd(inputs[0], self.dims)
            U *= self.adjoint_kernel
            np.copyto(outputs[0], ifftd(U, self.dims).real)

    def is_diag(self, freq=False):
        """Is the lin op diagonal (in the frequency domain)?
        """
        return freq and self.input_nodes[0].is_diag(freq)

    def get_diag(self, freq=False):
        """Returns the diagonal representation (A^TA)^(1/2).

        Parameters
        ----------
        freq : bool
            Is the diagonal representation in the frequency domain?
        Returns
        -------
        dict of variable to ndarray
            The diagonal operator acting on each variable.
        """
        assert freq
        var_diags = self.input_nodes[0].get_diag(freq)
        self_diag = np.reshape(self.forward_kernel, self.size)
        for var in var_diags.keys():
            var_diags[var] = var_diags[var]*self_diag
        return var_diags

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return np.max(np.abs(self.forward_kernel))*input_mags[0]



from .lin_op import LinOp
import numpy as np
import uuid

class Variable(LinOp):
    """A variable.
    """
    def __init__(self, shape):
        self.uuid = uuid.uuid1()
        self._value = None
        super(Variable, self).__init__([], shape)

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        np.copyto(outputs[0], inputs[0])

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        np.copyto(outputs[0], inputs[0])

    def variables(self):
        return [self]

    def is_diag(self, freq=False):
        """Is the lin op diagonal (in the frequency domain)?
        """
        return True

    def get_diag(self, freq=False):
        """Returns the diagonal representation (A^TA)^(1/2).

        Parameters
        ----------
        freq : bool
            Is the diagonal representation in the frequency domain?
        Returns
        -------
        dict of variable to ndarray
            The diagonal operator acting on each variable.
        """
        return {self: np.ones(self.size)}

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        """Assign a value to the variable.
        """
        self._value = val

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return 1.0



import numpy as np
import math

def equil(K, iters=50, gamma=1e-1, M=math.log(1e4)):
    """Computes diagonal D, E so that DKE is approximately equilibrated.
    """
    m = K.output_size
    n = K.input_size
    alpha, beta = get_alpha_beta(m, n)

    u = np.zeros(m)
    v = np.zeros(n)
    ubar = u.copy()
    vbar = v.copy()

    in_buf = np.zeros(n)
    out_buf = np.zeros(m)

    # Main loop.
    for t in range(1,iters+1):
        step_size = 2/(gamma*(t+1))
        # u grad estimate.
        s = np.random.choice([-1,1], size=n)
        K.forward(np.exp(v)*s, out_buf)
        u_grad = np.exp(2*u)*np.square(out_buf) - alpha**2 + gamma*u

        # v grad estimate.
        w = np.random.choice([-1,1], size=m)
        K.adjoint(np.exp(u)*w, in_buf)
        v_grad = np.exp(2*v)*np.square(in_buf) - beta**2 + gamma*v

        u = project(u - step_size*u_grad, M)
        v = project(v - step_size*v_grad, M)
        # Update averages.
        ubar = 2*u/(t+2) + t*ubar/(t+2)
        vbar = 2*v/(t+2) + t*vbar/(t+2)

    return np.exp(ubar), np.exp(vbar)

def get_alpha_beta(m,n):
    return (n/m)**(0.25), (m/n)**(0.25)

def project(x, M):
    """Project x onto [-M, M]^n.
    """
    return np.minimum(M, np.maximum(x, -M, out=x), out=x)

# Comparison method.
def f(A, u, v, gamma, p=2):
    m, n = A.shape
    alpha, beta = get_alpha_beta(m,n)
    total = (1./p)*np.exp(p*u).T.dot(np.power(np.abs(A), p)).dot(np.exp(p*v))
    total += -alpha**p*u.sum() - beta**p*v.sum() + (gamma/2)*((u*u).sum() + (v*v).sum())
    return np.sum(total)

def get_grad(A, u, v, gamma, p=2):
    m, n = A.shape
    alpha, beta = get_alpha_beta(m, n)

    tmp = np.diag(np.exp(p*u)).dot((A*A)).dot(np.exp(p*v))
    grad_u = tmp - alpha**p + gamma*u
    du = -grad_u/(2*tmp + gamma)

    tmp = np.diag(np.exp(p*v)).dot((A.T*A.T)).dot(np.exp(p*u))
    grad_v = tmp - beta**p + gamma*v
    dv = -grad_v/(2*tmp + gamma)

    return du, dv, grad_u, grad_v

def newton_equil(A, gamma, max_iters):
    alpha = 0.25
    beta = 0.5
    m, n = A.shape
    u = np.zeros(m)
    v = np.zeros(n)
    for i in range(max_iters):
        du, dv, grad_u, grad_v = get_grad(A, u, v, gamma)
        # Backtracking line search.
        t = 1
        obj = f(A, u, v, gamma)
        grad_term = np.sum(alpha*(grad_u.dot(du) + grad_v.dot(dv)))
        while True:
            new_obj = f(A, u + t*du, v + t*dv, gamma)
            if new_obj > obj + t*grad_term:
                t = beta*t
            else:
                u = u + t*du
                v = v + t*dv
                break
    return np.exp(u), np.exp(v)



from .diff_fn import diff_fn
from .nonneg import nonneg, weighted_nonneg, masked_nonneg
from .norm1 import norm1, weighted_norm1
from .sum_squares import sum_squares, weighted_sum_squares, least_squares, lsqr_options, cg_options
from .group_norm1 import group_norm1, weighted_group_norm1
from .poisson_norm import poisson_norm, weighted_poisson_norm
from .patch_NLM import patch_NLM
from .sum_entries import sum_entries, zero_prox
# from .matlab_external import matlab_external



import numpy as np

class Edge(object):
    """The edge between two lin ops.
    """
    def __init__(self, start, end, shape):
        self.start = start
        self.end = end
        self.shape = shape
        self.data = np.zeros(self.shape)
        self.mag = None # Used to get norm bounds.

    @property
    def size(self):
        return reduce(lambda x,y: x*y, self.shape, 1)



from .lin_op import LinOp
import numpy as np

class subsample(LinOp):
    """Samples every steps[i] pixel along axis i,
       starting with pixel 0.
    """
    def __init__(self, arg, steps):
        self.steps = steps
        shape = tuple([(dim-1)//step + 1 for dim,step in zip(arg.shape, steps)])
        super(subsample, self).__init__([arg], shape)

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        # Subsample.
        selection = self.get_selection()
        np.copyto(outputs[0], inputs[0][selection])

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        # Fill in with zeros.
        selection = self.get_selection()
        outputs[0][:] *= 0
        outputs[0][selection] = inputs[0]

    def get_selection(self):
        """Return a tuple of slices to index into numpy arrays.
        """
        selection = []
        for step in self.steps:
            selection.append(slice(None,None,step))
        return tuple(selection)

    def is_gram_diag(self, freq=False):
        """Is the lin op's Gram matrix diagonal (in the frequency domain)?
        """
        return not freq and self.input_nodes[0].is_diag(freq)

    def get_diag(self, freq=False):
        """Returns the diagonal representation (A^TA)^(1/2).

        Parameters
        ----------
        freq : bool
            Is the diagonal representation in the frequency domain?
        Returns
        -------
        dict of variable to ndarray
            The diagonal operator acting on each variable.
        """
        assert not freq
        var_diags = self.input_nodes[0].get_diag(freq)
        selection = self.get_selection()
        self_diag = np.zeros(self.input_nodes[0].shape)
        self_diag[selection] = 1
        for var in var_diags.keys():
            var_diags[var] = var_diags[var]*self_diag.ravel()
        return var_diags

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return input_mags[0]



from .lin_op import LinOp

def LinOpFactory(input_shape, output_shape, forward, adjoint, norm_bound=None):
    """Returns a function to generate a custom LinOp.

    Parameters
    ----------
    input_shape : tuple
        The dimensions of the input.
    output_shape : tuple
        The dimensions of the output.
    forward : function
        Applies the operator to an input array and writes to an output.
    adjoint : function
        Applies the adjoint operator to an input array and writes to an output.
    norm_bound : float, optional
        An upper bound on the spectral norm of the operator.
    """
    def get_black_box(arg):
        return BlackBox(arg, input_shape, output_shape,
                        forward, adjoint, norm_bound)
    return get_black_box

class BlackBox(LinOp):
    """A black-box lin op specified by the user.
    """
    def __init__(self, arg, input_shape, output_shape,
                 forward, adjoint, norm_bound=None):
        assert arg.shape == input_shape
        self._forward = forward
        self._adjoint = adjoint
        self._norm_bound = norm_bound
        super(BlackBox, self).__init__([arg], output_shape)

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        self._forward(inputs[0], outputs[0])

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        self._adjoint(inputs[0], outputs[0])

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        if self._norm_bound is None:
            return super(BlackBox, self).norm_bound(input_mags)
        else:
            return self._norm_bound*input_mags[0]



from proximal.tests.base_test import BaseTest
from proximal.prox_fns import *
from proximal.lin_ops import *
from proximal.algorithms import *
from proximal.algorithms.merge import merge_all
import numpy as np
import cvxpy as cvx
import proximal.algorithms.absorb

class TestTransforms(BaseTest):

    def test_absorb_lin_op(self):
        """Test absorb lin op operator.
        """
        # norm1.
        tmp = Variable(10)
        v = np.arange(10)*1.0 - 5.0

        fn = norm1(mul_elemwise(-v, tmp), alpha=5.)
        rho = 2
        new_prox = absorb_lin_op(fn)[0]
        x = new_prox.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, np.sign(v)*np.maximum(np.abs(v) - 5.*np.abs(v)/rho,0))

        fn = norm1(mul_elemwise(-v, mul_elemwise(2*v, tmp)), alpha=5.)
        rho = 2
        new_prox = absorb_lin_op(fn)[0]
        x = new_prox.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, np.sign(v)*np.maximum(np.abs(v) - 5.*np.abs(v)/rho,0))
        new_prox = absorb_lin_op(new_prox)[0]
        x = new_prox.prox(rho, v.copy())
        new_v = 2*v*v
        self.assertItemsAlmostEqual(x, np.sign(new_v)*np.maximum(np.abs(new_v) - 5.*np.abs(new_v)/rho,0))

        # nonneg.
        tmp = Variable(10)
        v = np.arange(10)*1.0 - 5.0

        fn = nonneg(mul_elemwise(-v, tmp), alpha=5.)
        rho = 2
        new_prox = absorb_lin_op(fn)[0]
        x = new_prox.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, fn.prox(rho, -np.abs(v)))

        # sum_squares.
        tmp = Variable(10)
        v = np.arange(10)*1.0 - 5.0

        alpha = 5.
        val = np.arange(10)
        fn = sum_squares(mul_elemwise(-v, tmp), alpha=alpha, c=val)
        rho = 2
        new_prox = absorb_lin_op(fn)[0]
        x = new_prox.prox(rho, v.copy())

        cvx_x = cvx.Variable(10)
        prob = cvx.Problem(cvx.Minimize(cvx.sum_squares(cvx_x - v)*(rho/2) + \
            5*cvx.sum_squares(cvx.mul_elemwise(-v, cvx_x)) + (val*-v).T*cvx_x
        ))
        prob.solve()
        self.assertItemsAlmostEqual(x, cvx_x.value, places=3)


        # Test scale.
        tmp = Variable(10)
        v = np.arange(10)*1.0 - 5.0

        fn = norm1(10*tmp)
        rho = 2
        new_prox = absorb_lin_op(fn)[0]
        x = new_prox.prox(rho, v.copy())
        cvx_x = cvx.Variable(10)
        prob = cvx.Problem(cvx.Minimize(cvx.sum_squares(cvx_x - v) + cvx.norm(10*cvx_x, 1)))
        prob.solve()
        self.assertItemsAlmostEqual(x, cvx_x.value, places=3)

        val = np.arange(10)
        fn = norm1(10*tmp, c=val, b=val, gamma=0.01)
        rho = 2
        new_prox = absorb_lin_op(fn)[0]
        x = new_prox.prox(rho, v.copy())
        cvx_x = cvx.Variable(10)
        prob = cvx.Problem(cvx.Minimize(cvx.sum_squares(cvx_x - v) + \
            cvx.norm(10*cvx_x - val, 1) + 10*val.T*cvx_x + cvx.sum_squares(cvx_x)
        ))
        prob.solve()
        self.assertItemsAlmostEqual(x, cvx_x.value, places=2)

        # sum_entries
        tmp = Variable(10)
        v = np.arange(10)*1.0 - 5.0

        fn = sum_entries(sum([10*tmp, mul_elemwise(v, tmp)]))

        funcs = absorb.absorb_all_lin_ops([fn])
        c = __builtins__['sum']([fn.c for fn in funcs])
        self.assertItemsAlmostEqual(c, v+10, places=3)

    def test_merge(self):
        """Test merging functions.
        """
        # sum_entries
        x = Variable(10)
        fn1 = sum_entries(x, gamma=1.0)
        fn2 = norm1(x)
        assert can_merge(fn1, fn2)
        merged = merge_fns(fn1, fn2)
        v = np.arange(10)*1.0 - 5.0
        prox_val1 = merged.prox(1.0, v.copy())
        tmp = norm1(x, c=np.ones(10), gamma=1.0)
        prox_val2 = tmp.prox(1.0, v.copy())
        self.assertItemsAlmostEqual(prox_val1, prox_val2)

        # sum_squares
        x = Variable(10)
        val = np.arange(10)
        fn1 = sum_squares(x, gamma=1.0, beta=2.0, alpha=3.0, b=val)
        fn2 = norm1(x)
        assert can_merge(fn1, fn2)
        merged = merge_fns(fn1, fn2)
        v = np.arange(10)*1.0 - 5.0
        prox_val1 = merged.prox(1.0, v.copy())
        tmp = norm1(x, c=-12*val, gamma=1.0+12, d=val.dot(val))
        prox_val2 = tmp.prox(1.0, v.copy())
        self.assertItemsAlmostEqual(prox_val1, prox_val2)

    def test_merge_all(self):
        """Test function to merge all prox operators possible.
        """
        # merge all
        x = Variable(10)
        lin_op = grad(x)
        fns = [sum_squares(lin_op), sum_entries(lin_op), nonneg(lin_op)]
        merged = merge_all(fns)
        assert len(merged) == 1
        v = np.reshape(np.arange(10)*1.0 - 5.0, (10,1))
        prox_val1 = merged[0].prox(1.0, v.copy())
        tmp = nonneg(lin_op, c=np.ones((10,1)), gamma=1.0)
        prox_val2 = tmp.prox(1.0, v.copy())
        self.assertItemsAlmostEqual(prox_val1, prox_val2)

    def test_const_val(self):
        """Test obtaining the constant offset.
        """
        x = Variable(10)
        b = np.arange(10)
        expr = x - b
        self.assertItemsAlmostEqual(-b, expr.get_offset())
        fn = sum_squares(expr)
        new_fn = absorb_offset(fn)
        self.assertItemsAlmostEqual(b, new_fn.b)



import timeit

class ConvergenceLog(object):
    """A log of the runtime for an operation.
    """
    def __init__(self):
        self.total_time = 0.0
        self.iter_time = []
        self.objective_val = [ ]
        self.lastticstamp = None

    def record_timing(self, elapsed):
        """Updates the log with the new time.
        """
        self.total_time += elapsed
        self.iter_time.append(self.total_time)

    def record_objective(self, val):
        """Updates the log with the new time.
        """
        self.objective_val.append(val)

    def tic(self):
        """ Default timer
        Example: t = tic()
             ... code
             elapsed = toc(t)
             print( '{0}: {1:.4f}ms'.format(message, elapsed) )
        """

        t = timeit.default_timer()
        self.lastticstamp = t
        return t

    def toc(self):
        """ See tic f
        """

        #Last tic
        if self.lastticstamp is None:
            raise Exception('Error: Call to toc did never call tic before.')
        else:
            t = self.lastticstamp
            #Measure time in ms
            elapsed = (timeit.default_timer() - t) * 1000.0  #in ms
            # Update recrod.
            self.record_timing(elapsed)
            self.lastticstamp = None
            return elapsed


from .lin_op import LinOp
import numpy as np
from proximal.utils.utils import *
from proximal.halide.halide import *

class mul_elemwise(LinOp):
    """Elementwise multiplication weight*X with a fixed constant.
    """
    def __init__(self, weight, arg, implem = None):
        assert arg.shape == weight.shape
        self.weight = weight
        shape = arg.shape

        #Halide temp
        if len(shape) in [2,3]:
            self.weight = np.asfortranarray( self.weight.astype(np.float32) )
            self.tmpout = np.zeros(arg.shape, dtype=np.float32, order='F')

        super(mul_elemwise, self).__init__([arg], shape, implem)

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        if self.implementation == Impl['halide'] and (len(self.shape) in [2,3]):

            #Halide implementation
            tmpin = np.asfortranarray( inputs[0].astype(np.float32) )
            Halide('A_mask.cpp').A_mask( tmpin, self.weight, self.tmpout ) #Call
            np.copyto(outputs[0], self.tmpout)

        else:
            #Numpy
            np.multiply(inputs[0], self.weight, outputs[0])

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        self.forward(inputs, outputs)

    def is_diag(self, freq=False):
        """Is the lin op diagonal (in the frequency domain)?
        """
        return not freq and self.input_nodes[0].is_diag(freq)

    def get_diag(self, freq=False):
        """Returns the diagonal representation (A^TA)^(1/2).

        Parameters
        ----------
        freq : bool
            Is the diagonal representation in the frequency domain?
        Returns
        -------
        dict of variable to ndarray
            The diagonal operator acting on each variable.
        """
        assert not freq
        var_diags = self.input_nodes[0].get_diag(freq)
        self_diag = np.reshape(self.weight, self.size)
        for var in var_diags.keys():
            var_diags[var] = var_diags[var]*self_diag
        return var_diags

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return np.max(np.abs(self.weight))*input_mags[0]



import numpy as np
import math
from proximal.lin_ops import CompGraph, scale, Variable, vstack
from proximal.prox_fns import least_squares
from proximal.utils.timings_log import TimingsLog, TimingsEntry
from invert import get_least_squares_inverse, get_diag_quads


def partition(prox_fns, try_diagonalize=True):
    """Divide the proxable functions into sets Psi and Omega.
    """
   # Merge these quadratic functions with the v update.
    quad_funcs = []
    # All lin ops must be gram diagonal for the least squares problem
    # to be diagonal.
    func_opts = {True:[], False:[]}
    for freq in [True, False]:
        if all([fn.lin_op.is_gram_diag(freq) for fn in prox_fns]):
            func_opts[freq] = get_diag_quads(prox_fns, freq)
    # Quad funcs is the max cardinality set.
    if len(func_opts[True]) >= len(func_opts[False]):
        quad_funcs = func_opts[True]
    else:
        quad_funcs = func_opts[False]
    psi_fns = [fn for fn in prox_fns if fn not in quad_funcs]
    return psi_fns, quad_funcs

def solve(psi_fns, omega_fns,
          rho_0 = 1.0, rho_scale = math.sqrt(2.0) * 2.0, rho_max = 2**8,
          max_iters = -1, max_inner_iters = 100, x0 = None,
          eps_rel=1e-3, eps_abs=1e-3,
          lin_solver="cg", lin_solver_options=None,
          try_diagonalize=True, scaled=False, try_fast_norm=False,
          metric=None, convlog=None, verbose=0):
    prox_fns = psi_fns + omega_fns
    stacked_ops = vstack([fn.lin_op for fn in psi_fns])
    K = CompGraph(stacked_ops)
    # Rescale so (1/2)||x - b||^2_2
    rescaling = np.sqrt(2.)
    quad_ops = []
    quad_weights = []
    const_terms = []
    for fn in omega_fns:
        fn = fn.absorb_params()
        quad_ops.append( scale(rescaling*fn.beta, fn.lin_op) )
        quad_weights.append( rescaling*fn.beta )
        const_terms.append( fn.b.flatten()*rescaling )

    # Get optimize inverse (tries spatial and frequency diagonalization)
    op_list = [fn.lin_op for fn in psi_fns] + quad_ops
    stacked_ops = vstack(op_list)
    x_update = get_least_squares_inverse(op_list, None,
                                        try_diagonalize, verbose)

    # Initialize
    if x0 is not None:
        x = np.reshape( x0, K.input_size )
    else:
        x = np.zeros(K.input_size)

    Kx = np.zeros(K.output_size)
    w = Kx.copy()

    Kxn = np.zeros(K.output_size)
    KTx = np.zeros(K.input_size)

    #Temporary iteration counts
    x_prev = x.copy()

    # Log for prox ops.
    prox_log = TimingsLog(prox_fns)
    # Time iterations.
    iter_timing = TimingsEntry("HQS iteration")
    inner_iter_timing = TimingsEntry("HQS inner iteration")
    #Convergence log for initial iterate
    if convlog != None:
        K.update_vars(x)
        objval = sum([fn.value for fn in prox_fns])
        convlog.record_objective( objval )
        convlog.record_timing( 0.0 )

    #Rho scedule
    rho = rho_0
    i = 0
    while rho < rho_max and i < max_iters:
        iter_timing.tic()
        if convlog != None:
            convlog.tic()

        #Update rho for quadratics
        for idx, op in enumerate(quad_ops):
            op.scalar = quad_weights[idx] / np.sqrt(rho)
        x_update = get_least_squares_inverse(op_list, CompGraph(stacked_ops),
                                             try_diagonalize, verbose)

        for ii in range(max_inner_iters) :
            inner_iter_timing.tic()
            # Update Kx.
            K.forward(x, Kx)

            # Prox update to get w.
            offset = 0
            w_prev = w.copy()
            for fn in psi_fns:
                slc = slice(offset, offset+fn.lin_op.size, None)
                # Apply and time prox.
                prox_log[fn].tic()
                w[slc] = fn.prox( rho, np.reshape(Kx[slc], fn.lin_op.shape), ii ).flatten()
                prox_log[fn].toc()
                offset += fn.lin_op.size

            # Update x.
            x_prev[:] = x
            tmp = np.hstack([w] +  [cterm/np.sqrt(rho) for cterm in const_terms] )
            x = x_update.solve(tmp, x_init=x, lin_solver=lin_solver, options=lin_solver_options)

            # Very basic convergence check.
            r_x = np.linalg.norm(x_prev - x)
            eps_x = eps_rel * np.prod(K.input_size)

            r_w = np.linalg.norm(w_prev - w)
            eps_w = eps_rel * np.prod(K.output_size)

            #Convergence log
            if convlog != None:
                convlog.toc()
                K.update_vars(x)
                objval = sum([fn.value for fn in prox_fns])
                convlog.record_objective( objval )

            #Show progess
            if verbose > 0:
                #Evaluate objective only if required (expensive !)
                objstr = ''
                if verbose == 2:
                    K.update_vars(x)
                    objstr = ", obj_val = %02.03e" % sum([fn.value for fn in prox_fns])

                #Evaluate metric potentially
                metstr = '' if metric is None else ", {}".format( metric.message(x) )
                print "iter [%02d (rho=%2.1e) || %02d]:" \
                      "||w - w_prev||_2 = %02.02e (eps=%02.03e)" \
                      "||x - x_prev||_2 = %02.02e (eps=%02.03e)%s%s" \
                        % (i, rho, ii, r_x, eps_x, r_w, eps_w, objstr, metstr)

            inner_iter_timing.toc()
            if r_x < eps_x and r_w < eps_w:
                break

        #Update rho
        rho = np.minimum( rho * rho_scale, rho_max )
        i += 1
        iter_timing.toc()

    # Print out timings info.
    if verbose > 0:
        print iter_timing
        print inner_iter_timing
        print "prox funcs:"
        print prox_log
        print "K forward ops:"
        print K.forward_log
        print "K adjoint ops:"
        print K.adjoint_log

    # Assign values to variables.
    K.update_vars(x)

    # Return optimal value.
    return sum([fn.value for fn in prox_fns])



from .prox_fn import ProxFn
import numpy as np
import matlab.engine

class matlab_external(ProxFn):
    """The function for matlab prior
    """
    def __init__(self, lin_op, matengine, proxfunc, evalfunc=None, params=None, **kwargs):

        self.matengine = matengine
        self.params = params
        self.proxfunc = proxfunc
        self.evalfunc = evalfunc

        super(matlab_external, self).__init__(lin_op, **kwargs)

    def _prox(self, rho, v, *args, **kwargs):
        """ calls matlab proximal function on v
        """
        assert it is not None

        #Run external matlab method
        proxmethod = getattr(self.matengine, self.proxfunc)

        vmat = matlab.double(v.tolist())
        if self.params is None:
            pres = np.array( proxmethod(vmat, rho, it) )
        else:
            paramsmat = matlab.double(self.params.tolist())
            pres = np.array( proxmethod(vmat, rho, it, paramsmat) )

        np.copyto( v, pres )
        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """

        if self.evalfunc is None:
            return np.inf
        else:
            evalmethod = getattr(self.matengine, self.evalfunc)

            vmat = matlab.double(v)
            if self.params is None:
                peval = np.float32( evalmethod(vmat) )
            else:
                paramsmat = matlab.double(self.params.tolist())
                peval = np.float32( evalmethod(vmat, paramsmat) )

            return peval

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.matengine, self.proxfunc, self.evalfunc, self.params]




from __future__ import division
from proximal.tests.base_test import BaseTest
from proximal.lin_ops import *
from proximal.halide.halide import *
from proximal.utils.utils import *
import numpy as np
import scipy as sp

import os
from PIL import Image
from scipy import signal
from scipy import ndimage
import cv2

class TestLinOps(BaseTest):

    def test_subsample(self):
        """Test subsample lin op.
        """
        # Forward.
        var = Variable((10,5))
        fn = subsample(var, (2,1))
        x = np.arange(50)*1.0
        x = np.reshape(x, (10, 5))
        out = np.zeros(fn.shape)
        fn.forward([x], [out])
        self.assertItemsAlmostEqual(out, x[0::2,:])

        # Adjoint.
        x = np.arange(25)*1.0
        x = np.reshape(x, (5, 5))
        out = np.zeros(var.shape)
        fn.adjoint([x], [out])
        zeroed_x = np.zeros((10,5))*1.0
        zeroed_x[0::2,:] = x
        self.assertItemsAlmostEqual(out, zeroed_x)

        # Constant arg.
        val = np.arange(50)*1.0
        val = np.reshape(val, (10, 5))
        fn = subsample(val, (2,1))
        self.assertItemsAlmostEqual(fn.value, val[0::2,:])

        # Diagonal form.
        var = Variable((5,1))
        fn = subsample(var, (2,1))
        assert not fn.is_gram_diag(freq=True)
        assert fn.is_gram_diag(freq=False)
        print fn.get_diag(freq=False)[var]
        self.assertItemsAlmostEqual(fn.get_diag(freq=False)[var],
            [1,0,1,0,1])

    def test_sum(self):
        # Forward.
        x = Variable((2,3))
        y = Variable((2,3))
        fn = sum([x, y])
        x_val = np.reshape(np.arange(6)*1.0, x.shape)
        y_val = np.reshape(np.arange(6)*1.0 - 5, x.shape)
        out = np.zeros(fn.shape)
        fn.forward([x_val, y_val], [out])
        self.assertItemsAlmostEqual(out, 2*np.arange(6) - 5)

        # Adjoint.
        x_val = np.reshape(np.arange(6)*1.0, x.shape)
        out = [np.zeros(fn.shape), np.zeros(fn.shape)]
        fn.adjoint([x_val], out)
        for arr in out:
            self.assertItemsAlmostEqual(arr, x_val)

        # Constant args.
        x = Variable((2,3))
        y = Variable((2,3))
        x_val = np.reshape(np.arange(6)*1.0, x.shape)
        y_val = np.reshape(np.arange(6)*1.0 - 5, x.shape)
        fn = sum([x_val, y_val])
        self.assertItemsAlmostEqual(fn.value, 2*np.arange(6) - 5)

        # Diagonal form.
        x = Variable(5)
        term = mul_elemwise(np.arange(5)-3, x)
        fn = sum([term, x])
        assert not fn.is_diag(freq=True)
        assert fn.is_diag(freq=False)
        self.assertItemsAlmostEqual(fn.get_diag(freq=False)[x],
            np.arange(5)-3 + np.ones(5))

    def test_conv(self):
        """Test convolution lin op.
        """
        # Forward.
        var = Variable((2,3))
        kernel = np.array([[1,2,3], [4,5,6]]) # 2x3
        fn = conv(kernel, var)
        x = np.arange(6)*1.0
        x = np.reshape(x,(2, 3))
        out = np.zeros(fn.shape)

        fn.forward([x], [out])
        ks = np.array(kernel.shape)
        cc = np.floor( (ks - 1) / 2.0 ).astype(np.int)  #Center coordinate
        y = np.zeros((2,3))
        for i in range(2):
            for j in range(3):
                for s in range(2):
                    for t in range(3):
                        y[i,j] += kernel[ks[0]-1-s, ks[1]-1-t] * x[(s + i - cc[0]) % 2,(t + j - cc[1]) % 3]

        #For loop same as convolve
        #y = ndimage.convolve(x, kernel, mode='wrap')

        self.assertItemsAlmostEqual(out, y)

        # Adjoint.#
        x = np.arange(6)*1.0
        x = np.reshape(x, (2, 3))
        out = np.zeros(var.shape)
        fn.adjoint([x], [out])
        y = np.zeros((2,3))
        for i in range(2):
            for j in range(3):
                for s in range(2):
                    for t in range(3):
                        y[i,j] += kernel[s, t] * x[(s + i - ( ks[0]-1-cc[0]) )  % 2,(t + j - (ks[1]-1-cc[1]) ) % 3]

        #For loop same as correlate
        #y = ndimage.correlate(x, kernel, mode='wrap')

        self.assertItemsAlmostEqual(out, y)

        # Diagonal form.
        x = Variable(5)
        kernel = np.arange(5)
        fn = conv(kernel, x)
        assert fn.is_diag(freq=True)
        assert not fn.is_diag(freq=False)
        forward_kernel = psf2otf(kernel, (5,), 1)
        self.assertItemsAlmostEqual(fn.get_diag(freq=True)[x],
            forward_kernel)

    def test_conv_halide(self):
        """Test convolution lin op in halide.
        """

        #Load image
        testimg_filename = os.path.join( os.path.dirname(os.path.realpath(__file__)), 'data', 'angela.jpg' )
        img = Image.open(testimg_filename)  # opens the file using Pillow - it's not an array yet
        np_img = np.asfortranarray( im2nparray(img) )

        #Convert to gray
        np_img = np.mean( np_img, axis=2)

        #Test problem
        output = np.zeros_like(np_img);
        K = np.ones((11,11), dtype=np.float32, order='FORTRAN')
        K /= np.prod( K.shape )

        #Convolve in halide
        Halide('A_conv.cpp').A_conv(np_img, K, output)

        #Convolve in scipy
        output_ref = signal.convolve2d(np_img,K, mode='same', boundary='wrap')

        #Transpose
        output_corr = np.zeros_like(np_img);
        Halide('At_conv.cpp').At_conv(np_img, K, output_corr) #Call

        output_corr_ref = signal.convolve2d(np_img, np.flipud(np.fliplr(K)), mode='same', boundary='wrap')

        self.assertItemsAlmostEqual(output, output_ref)
        self.assertItemsAlmostEqual(output_corr, output_corr_ref)

    def vstack(self):
        """Test vstack operator.
        """
        # diagonals.
        x = Variable(1)
        y = Variable(1)
        fn = vstack([x,y])
        assert fn.is_gram_diag(freq=False)
        assert fn.is_gram_diag(freq=True)
        fn = vstack([fn])
        assert fn.is_gram_diag(freq=False)
        assert fn.is_gram_diag(freq=True)

    def test_combo(self):
        """Test subsampling followed by convolution.
        """
        # Forward.
        var = Variable((2,3))
        kernel = np.array([[1,2,3]]) # 2x3
        fn = vstack([conv(kernel, subsample(var, (2,1)))])
        fn = CompGraph(fn)
        x = np.arange(6)*1.0
        x = np.reshape(x, (2, 3))
        out = np.zeros(fn.output_size)
        fn.forward(x.flatten(), out)
        y = np.zeros((1,3))

        xsub = x[::2,::1]
        y = ndimage.convolve(xsub, kernel, mode='wrap')

        self.assertItemsAlmostEqual(np.reshape(out, y.shape), y)

        # Adjoint.
        x = np.arange(3)*1.0
        x = np.reshape(x, (1, 3))
        out = np.zeros(var.size)
        fn.adjoint(x.flatten(), out)

        y = ndimage.correlate(x, kernel, mode='wrap')
        y2 = np.zeros((2,3))
        y2[::2,:] = y

        self.assertItemsAlmostEqual(np.reshape(out, y2.shape), y2)
        out = np.zeros(var.size)
        fn.adjoint(x.flatten(), out)
        self.assertItemsAlmostEqual(np.reshape(out, y2.shape), y2)

    def test_mul_elemwise(self):
        """Test mul_elemwise lin op.
        """
        # Forward.
        var = Variable((2,5))
        W = np.arange(10)
        W = np.reshape(W, (2,5))
        fn = mul_elemwise(W, var)
        x = W.copy()
        out = np.zeros(x.shape)
        fn.forward([x], [out])
        self.assertItemsAlmostEqual(out, W*W)

        # Adjoint.
        x = W.copy()
        out = np.zeros(x.shape)
        fn.adjoint([x], [out])
        self.assertItemsAlmostEqual(out, W*W)

        # Diagonal form.
        x = Variable(5)
        fn = mul_elemwise(np.arange(5)-3, x)
        assert not fn.is_diag(freq=True)
        assert fn.is_diag(freq=False)
        self.assertItemsAlmostEqual(fn.get_diag(freq=False)[x],
            np.arange(5)-3)

    def test_black_box(self):
        """Test custom linear operators.
        """
        scale = 2
        def op(input, output):
            output[:] = 2*input

        my_op = LinOpFactory((2,2), (2,2), op, op, scale)
        x = Variable((2,2))
        expr = my_op(x)
        val = np.reshape(np.arange(4), (2,2))
        x.value = val
        self.assertItemsAlmostEqual(expr.value, 2*val)
        output = np.zeros((2,2))
        expr.forward([val], [output])
        self.assertItemsAlmostEqual(output, 2*val)
        expr.adjoint([val], [output])
        self.assertItemsAlmostEqual(output, 2*val)

    def test_op_overloading(self):
        """Test operator overloading.
        """
        # Multiplying by a scalar.

        # Forward.
        var = Variable((2,5))
        W = np.arange(10)
        W = np.reshape(W, (2,5))
        fn = -2*mul_elemwise(W, var)
        fn = CompGraph(fn)
        x = W.copy()
        out = np.zeros(x.shape)
        fn.forward(x.flatten(), out)
        self.assertItemsAlmostEqual(out, -2*W*W)

        # Adjoint.
        x = W.copy()
        out = np.zeros(x.shape).flatten()
        fn.adjoint(x, out)
        self.assertItemsAlmostEqual(out, -2*W*W)

        # Forward.
        var = Variable((2,5))
        W = np.arange(10)
        W = np.reshape(W, (2,5))
        fn = mul_elemwise(W, var)*0.5
        fn = CompGraph(fn)
        x = W.copy()
        out = np.zeros(x.shape)
        fn.forward(x.flatten(), out)
        self.assertItemsAlmostEqual(out, W*W/2.)

        # Adjoint.
        x = W.copy()
        out = np.zeros(x.shape).flatten()
        fn.adjoint(x, out)
        self.assertItemsAlmostEqual(out, W*W/2.)


        # Dividing by a scalar.
        # Forward.
        var = Variable((2,5))
        W = np.arange(10)
        W = np.reshape(W, (2,5))
        fn = mul_elemwise(W, var)/2
        fn = CompGraph(fn)
        x = W.copy()
        out = np.zeros(x.shape)
        fn.forward(x, out)
        self.assertItemsAlmostEqual(out, W*W/2.)

        # Adding lin ops.
        # Forward.
        x = Variable((2,5))
        W = np.arange(10)
        W = np.reshape(W, (2,5))
        fn = mul_elemwise(W, x)
        fn = fn + x + x
        self.assertEquals(len(fn.input_nodes), 3)
        fn = CompGraph(fn)
        x = W.copy()
        out = np.zeros(fn.shape)
        fn.forward(x, out)
        self.assertItemsAlmostEqual(out, W*W + 2*W)

        # Adjoint.
        x = W.copy()
        out = np.zeros(x.shape).flatten()
        fn.adjoint(x, out)
        self.assertItemsAlmostEqual(out, W*W + 2*W)

        # Adding in a constant.
        # CompGraph should ignore the constant.
        x = Variable((2,5))
        W = np.arange(10)
        W = np.reshape(W, (2,5))
        fn = mul_elemwise(W, x)
        fn = fn + x + W 
        self.assertEquals(len(fn.input_nodes), 3)
        fn = CompGraph(fn)
        x = W.copy()
        out = np.zeros(fn.shape)
        fn.forward(x, out)
        self.assertItemsAlmostEqual(out, W*W + W)

       # Subtracting lin ops.
        # Forward.
        x = Variable((2,5))
        W = np.arange(10)
        W = np.reshape(W, (2,5))
        fn = -mul_elemwise(W, x)
        fn = x + x - fn
        self.assertEquals(len(fn.input_nodes), 3)
        fn = CompGraph(fn)
        x = W.copy()
        out = np.zeros(fn.shape)
        fn.forward(x, out)
        self.assertItemsAlmostEqual(out, W*W + 2*W)

        # Adjoint.
        x = W.copy()
        out = np.zeros(x.shape).flatten()
        fn.adjoint(x, out)
        self.assertItemsAlmostEqual(out, W*W + 2*W)

    def test_mask_halide(self):
        """Test mask lin op in halide.
        """

        #Load image
        testimg_filename = os.path.join( os.path.dirname(os.path.realpath(__file__)), 'data', 'angela.jpg' )
        img = Image.open(testimg_filename)  # opens the file using Pillow - it's not an array yet
        np_img = np.asfortranarray( im2nparray(img) )

        #Test problem
        output = np.zeros_like(np_img);
        mask = np.asfortranarray(  np.random.randn( *list(np_img.shape) ).astype(np.float32) )
        mask = np.maximum( mask, 0.)

        Halide('A_mask.cpp').A_mask(np_img, mask, output) #Call
        output_ref = mask * np_img

        #Transpose
        output_trans = np.zeros_like(np_img);
        Halide('At_mask.cpp').At_mask(np_img, mask, output_trans) #Call

        self.assertItemsAlmostEqual(output, output_ref)
        self.assertItemsAlmostEqual(output_trans, output_ref)

    def test_grad_halide(self):
        """Test gradient lin op in halide.
        """

        #Load image
        testimg_filename = os.path.join( os.path.dirname(os.path.realpath(__file__)), 'data', 'angela.jpg' )
        img = Image.open(testimg_filename)  # opens the file using Pillow - it's not an array yet
        np_img = np.asfortranarray( im2nparray(img) )

        #Convert to gray
        np_img = np.mean( np_img, axis=2)

        #Test problem
        output = np.zeros( (np_img.shape[0], np_img.shape[1], np_img.shape[2] if (len(np_img.shape) > 2) else 1, 2) , dtype=np.float32, order='FORTRAN');

        #Gradient in halide
        Halide('A_grad.cpp').A_grad(np_img, output) #Call

        #Compute comparison
        f = np_img
        if len(np_img.shape) == 2:
            f = f[..., np.newaxis]

        ss = f.shape;
        fx = f[ :, np.r_[1:ss[1],ss[1] - 1],:] - f;
        fy = f[ np.r_[1:ss[0],ss[0] - 1],:,:] - f;
        Kf = np.asfortranarray( np.stack( (fy, fx), axis=-1 ) )

        #Transpose
        output_trans = np.zeros(f.shape, dtype=np.float32, order='F');
        Halide('At_grad.cpp').At_grad(Kf, output_trans) #Call

        #Compute comparison (Negative divergence)
        Kfy = Kf[:,:,:,0]
        fy = Kfy - Kfy[ np.r_[0,0:ss[0]-1],:,:];
        fy[0,:,:] = Kfy[0,:,:];
        fy[-1,:,:] = -Kfy[-2,:,:];

        Kfx = Kf[:,:,:,1]
        ss = Kfx.shape
        fx = Kfx - Kfx[ :, np.r_[0,0:ss[1]-1],:]
        fx[:,0,:] = Kfx[:,0,:];
        fx[:,-1,:] = -Kfx[:,-2,:];

        KtKf = -fx -fy
        # TODO are these wrong?
        # self.assertItemsAlmostEqual(output, Kf)
        # self.assertItemsAlmostEqual(output_trans, KtKf)

    def test_warp_halide(self):
        """Test warp lin op in halide.
        """

        #Load image
        testimg_filename = os.path.join( os.path.dirname(os.path.realpath(__file__)), 'data', 'angela.jpg' )
        img = Image.open(testimg_filename)  # opens the file using Pillow - it's not an array yet
        np_img = np.asfortranarray( im2nparray(img) )

        #Convert to gray
        np_img = np.mean( np_img, axis=2)

        #Generate problem
        theta_rad = 5.0 * np.pi / 180.0;
        H = np.array([[np.cos(theta_rad),-np.sin(theta_rad), 0.0001 ],\
                      [np.sin(theta_rad), np.cos(theta_rad), 0.0003 ],\
                      [0.         , 0.           , 1.    ] ], dtype=np.float32, order='F')

        #Reference
        output_ref = cv2.warpPerspective(np_img, H.T, np_img.shape[1::-1], flags=cv2.INTER_LINEAR,
                                    borderMode=cv2.BORDER_CONSTANT, borderValue=0.) #cv2.WARP_INVERSE_MAP,

        #Halide
        output = np.zeros_like(np_img);
        Hc = np.asfortranarray( np.linalg.pinv(H)[...,np.newaxis] ) #Third axis for halide
        Halide('A_warp.cpp').A_warp(np_img, Hc, output) #Call

        #Transpose
        output_trans = np.zeros_like(np_img);
        Hinvc = np.asfortranarray( H[...,np.newaxis] ) #Third axis for halide
        Halide('At_warp.cpp').At_warp(output, Hinvc, output_trans) #Call

        #Compute reference
        output_ref_trans = cv2.warpPerspective(output_ref, H.T, np_img.shape[1::-1], flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP,
                                    borderMode=cv2.BORDER_CONSTANT, borderValue=0.)

        self.assertItemsAlmostEqual(output, output_ref,  places=1) #Opencv does inverse warp
        self.assertItemsAlmostEqual(output_trans, output_ref_trans,  places=1) #Opencv does inverse warp



from .prox_fn import ProxFn
import numpy as np
from proximal.utils.utils import *
from proximal.halide.halide import *

class poisson_norm(ProxFn):
    """The function ||x||_poisson(b) =
       x + Ind_+(x) - b * log( x_+ )
       This is the log likelihood of a poisson fit
    """
    def __init__(self, lin_op, bp, **kwargs):
        """Initialize temporary variables for _prox method.
        """
        self.bp = bp
        self.bph = np.asfortranarray( bp.astype(np.float32) )
        self.maskh = np.ones(lin_op.shape, dtype=np.float32, order='F')
        self.tmpout = np.zeros(lin_op.shape, dtype=np.float32, order='F')

        super(poisson_norm, self).__init__(lin_op, **kwargs)

    def _prox(self, rho, v, *args, **kwargs):
        """x = 1/2* ( v - 1./rho + sqrt( ||v - 1./rho||^2 + 4 * 1./rho * b ) )
        """
        if self.implementation == Impl['halide'] and (len(self.lin_op.shape) in [2,3,4]):

            #Halide implementation
            tmpin = np.asfortranarray(v)
            Halide('prox_Poisson.cpp').prox_Poisson(tmpin, self.maskh, self.bph, 1./rho, self.tmpout)
            np.copyto(v, self.tmpout)
        else:
            v = 0.5 * ( v - 1./rho + np.sqrt( (v - 1./rho)*(v - 1./rho) + 4. * 1./rho * self.bp ) )

        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """

        #Positivity penalty
        if np.min(v) < -1e-3:
            return np.inf

        #Other penalties
        vsum = v.copy();
        vsum -= self.bp * np.log( np.maximum(v, 1e-9) )
        return vsum.sum()

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.bp]

class weighted_poisson_norm(poisson_norm):
    """The function ||W.*x||_poisson.
    """
    def __init__(self, lin_op, bp, weight, **kwargs):
        self.weight = weight
        super(weighted_poisson_norm, self).__init__(lin_op, bp, **kwargs)

    def _prox(self, rho, v, *args, **kwargs):
        """x = 1/2* ( v - |W|/rho + sqrt( ||v - |W|/rho||^2 + 4 * 1./rho * b ) )
        """
        output = 0.5 * ( v - np.absolute(self.weight)/rho + np.sqrt( (v - np.absolute(self.weight)/rho)*(v - np.absolute(self.weight)/rho) + 4. * 1./rho * self.bp ) )

        #Reference
        idxs = self.weight == 0
        output[idxs] = v[idxs]
        return output

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return super(weighted_poisson_norm, self)._eval(self.weight*v)

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.bp, self.weight]



from .black_box import LinOpFactory
from .conv import conv
from .constant import Constant
from .comp_graph import CompGraph, est_CompGraph_norm
from .mul_elemwise import mul_elemwise
from .scale import scale
from .subsample import subsample
from .sum import sum, copy
from .variable import Variable
from .vstack import vstack, split
from .hstack import hstack
from .grad import grad
from .warp import warp
from .mul_color import mul_color
from .reshape import reshape



from proximal.tests.base_test import BaseTest
from proximal.prox_fns import *
from proximal.lin_ops import Variable
from proximal.halide.halide import *
from proximal.utils.utils import *
import cvxpy as cvx
import numpy as np
import scipy as sp

import os
from PIL import Image

class TestProxFn(BaseTest):

    def test_sum_squares(self):
        """Test sum squares prox fn.
        """
        # No modifiers.
        tmp = Variable(10)
        fn = sum_squares(tmp)
        rho = 1
        v = np.arange(10)*1.0
        x = fn.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, v*rho/(2+rho))

        rho = 2
        x = fn.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, v*rho/(2+rho))

        # With modifiers.
        mod_fn = sum_squares(tmp, alpha=2, beta=-1,
                             c=np.ones(10)*1.0, b=np.ones(10)*1.0, gamma=1)

        rho = 2
        v = np.arange(10)*1.0
        x = mod_fn.prox(rho, v.copy())

        # vhat = mod_fn.beta*(v - mod_fn.c/rho)*rho/(rho+2*mod_fn.gamma) - mod_fn.b
        # rho_hat = rho/(mod_fn.alpha*np.sqrt(np.abs(mod_fn.beta)))
        # xhat = fn.prox(rho_hat, vhat)
        x_var = cvx.Variable(10)
        cost = 2*cvx.sum_squares(-x_var - np.ones(10)) + np.ones(10).T*x_var + cvx.sum_squares(x_var) + (rho/2)*cvx.sum_squares(x_var - v)
        prob = cvx.Problem(cvx.Minimize(cost))
        prob.solve()

        self.assertItemsAlmostEqual(x, x_var.value, places=3)

    def test_norm1(self):
        """Test L1 norm prox fn.
        """
        # No modifiers.
        tmp = Variable(10)
        fn = norm1(tmp)
        rho = 1
        v = np.arange(10)*1.0 - 5.0
        x = fn.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, np.sign(v)*np.maximum(np.abs(v) - 1.0/rho,0))

        rho = 2
        x = fn.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, np.sign(v)*np.maximum(np.abs(v) - 1.0/rho,0))

        # With modifiers.
        mod_fn = norm1(tmp, alpha=0.1, beta=5,
                       c=np.ones(10)*1.0, b=np.ones(10)*-1.0, gamma=4)

        rho = 2
        v = np.arange(10)*1.0
        x = mod_fn.prox(rho, v.copy())

        # vhat = mod_fn.beta*(v - mod_fn.c/rho)*rho/(rho+2*mod_fn.gamma) - mod_fn.b
        # rho_hat = rho/(mod_fn.alpha*mod_fn.beta**2)
        # xhat = fn.prox(rho_hat, vhat)
        x_var = cvx.Variable(10)
        cost = 0.1*cvx.norm1(5*x_var + np.ones(10)) + np.ones(10).T*x_var + 4*cvx.sum_squares(x_var) + (rho/2)*cvx.sum_squares(x_var - v)
        prob = cvx.Problem(cvx.Minimize(cost))
        prob.solve()

        self.assertItemsAlmostEqual(x, x_var.value, places=3)

        # With weights.
        tmp = Variable(10)
        v = np.arange(10)*1.0 - 5.0
        fn = weighted_norm1(tmp, -v+1)
        rho = 2
        x = fn.prox(rho, v)
        self.assertItemsAlmostEqual(x, np.sign(v)*np.maximum(np.abs(v) - np.abs(-v+1)/rho,0))

    def test_nonneg(self):
        """Test I(x >= 0) prox fn.
        """
        # No modifiers.
        tmp = Variable(10)
        fn = nonneg(tmp)
        rho = 1
        v = np.arange(10)*1.0 - 5.0
        x = fn.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, np.maximum(v,0))

        rho = 2
        x = fn.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, np.maximum(v,0))

        # With modifiers.
        mod_fn = nonneg(tmp, alpha=0.1, beta=5,
                        c=np.ones(10)*1.0, b=np.ones(10)*-1.0, gamma=4)

        rho = 2
        v = np.arange(10)*1.0
        x = mod_fn.prox(rho, v.copy())

        vhat = mod_fn.beta*(v - mod_fn.c/rho)*rho/(rho+2*mod_fn.gamma) - mod_fn.b
        rho_hat = rho/(mod_fn.alpha*np.sqrt(np.abs(mod_fn.beta)))
        xhat = fn.prox(rho_hat, vhat)

        self.assertItemsAlmostEqual(x, (xhat + mod_fn.b)/mod_fn.beta)

        # With weights.
        tmp = Variable(10)
        v = np.arange(10)*1.0 - 5.0
        fn = weighted_nonneg(tmp, -v-4)
        rho = 2
        new_v = v.copy()
        idxs = (-v-4 != 0)
        new_v[idxs] = np.maximum((-v-4)[idxs]*v[idxs], 0.)/(-v-4)[idxs]
        x = fn.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, new_v)

    def test_norm1_halide(self):
        """Halide Norm 1 test
        """

        #Load image
        testimg_filename = os.path.join( os.path.dirname(os.path.realpath(__file__)), 'data', 'angela.jpg' )
        img = Image.open(testimg_filename)  # opens the file using Pillow - it's not an array yet
        np_img = np.asfortranarray( im2nparray(img) )

        #Convert to gray
        np_img = np.mean( np_img, axis=2)

        #Test problem
        v = np_img
        theta = 0.5

        #Output
        output = np.zeros_like(np_img)
        Halide('prox_L1.cpp').prox_L1(v, theta, output) #Call

        #Reference
        output_ref = np.maximum( 0.0, v - theta ) - np.maximum( 0.0, -v - theta )

        self.assertItemsAlmostEqual(output, output_ref)

    def test_isonorm1_halide(self):
        """Halide Norm 1 test
        """

        #Load image
        testimg_filename = os.path.join( os.path.dirname(os.path.realpath(__file__)), 'data', 'angela.jpg' )
        img = Image.open(testimg_filename)  # opens the file using Pillow - it's not an array yet
        np_img = np.asfortranarray( im2nparray(img) )

        #Convert to gray
        np_img = np.mean( np_img, axis=2)

        #Test problem
        theta = 0.5

        f = np_img
        if len(np_img.shape) == 2:
            f = f[..., np.newaxis]

        ss = f.shape;
        fx = f[ :, np.r_[1:ss[1],ss[1] - 1],:] - f;
        fy = f[ np.r_[1:ss[0],ss[0] - 1],:,:] - f;
        v = np.asfortranarray( np.stack( (fx, fy), axis=-1 ) )

        #Output
        output = np.zeros_like(v)
        Halide('prox_IsoL1.cpp').prox_IsoL1(v, theta, output) #Call

        #Reference
        normv = np.sqrt( np.multiply(v[:,:,:,0], v[:,:,:,0]) + np.multiply(v[:,:,:,1], v[:,:,:,1]) )
        normv = np.stack((normv,normv), axis =-1)
        with np.errstate(divide='ignore'):
            output_ref = np.maximum( 0.0, 1.0 - theta / normv ) * v

        self.assertItemsAlmostEqual(output, output_ref)

    def test_poisson_halide(self):
        """Halide Poisson norm test
        """

        #Load image
        testimg_filename = os.path.join( os.path.dirname(os.path.realpath(__file__)), 'data', 'angela.jpg' )
        img = Image.open(testimg_filename)  # opens the file using Pillow - it's not an array yet
        np_img = np.asfortranarray( im2nparray(img) )

        #Convert to gray
        np_img = np.mean( np_img, axis=2)

        #Test problem
        v = np_img
        theta = 0.5

        mask = np.asfortranarray(  np.random.randn( *list(np_img.shape) ).astype(np.float32) )
        mask = np.maximum( mask, 0.)
        b = np_img * np_img

        #Output
        output = np.zeros_like(v)

        tic()
        Halide('prox_Poisson.cpp').prox_Poisson(v, mask, b, theta, output) #Call
        print( 'Running took: {0:.1f}ms'.format( toc() ) )

        #Reference
        output_ref = 0.5 * ( v - theta + np.sqrt( (v - theta)*(v - theta) + 4 * theta * b ) )
        output_ref[mask <= 0.5] = v[mask <= 0.5]

        self.assertItemsAlmostEqual(output, output_ref)

    def test_diff_fn(self):
        """Test generic differentiable function operator.
        """
        # Least squares.
        tmp = Variable(10)
        fn = diff_fn(tmp, lambda x: np.square(x).sum(), lambda x: 2*x)
        rho = 1
        v = np.arange(10)*1.0 - 5.0
        x = fn.prox(rho, v.copy())
        self.assertItemsAlmostEqual(x, v/(2+rho))

        # -log
        n = 5
        tmp = Variable(n)
        fn = diff_fn(tmp, lambda x: -np.log(x).sum(), lambda x: -1.0/x)
        rho = 2
        v = np.arange(n)*2.0 + 1
        x = fn.prox(rho, v.copy())
        val = (v + np.sqrt(v**2 + 4/rho))/2
        self.assertItemsAlmostEqual(x, val)

    def test_sum_entries(self):
        """Sum of entries of lin op.
        """
        tmp = Variable(10)
        v = np.arange(10)*1.0 - 5.0
        fn = sum_entries(tmp)
        x = fn.prox(1.0, v.copy())
        self.assertItemsAlmostEqual(x, v - 1)

    def test_overloading(self):
        """Test operator overloading.
        """
        x = Variable(1)
        fn = sum_squares(x, b=1)
        val = fn.prox(2.0, 0)
        self.assertItemsAlmostEqual([val], [0.5])

        fn = 2*sum_squares(x, b=1)
        val = fn.prox(4.0, 0)
        self.assertItemsAlmostEqual([val], [0.5])

        fn = sum_squares(x, b=1)*2
        val = fn.prox(4.0, 0)
        self.assertItemsAlmostEqual([val], [0.5])

        fn = sum_squares(x, b=1)/2
        val = fn.prox(1.0, 0)
        self.assertItemsAlmostEqual([val], [0.5])

        fn1 = sum_squares(x, b=1)
        fn2 = norm1(x)
        arr = fn1 + fn2
        self.assertEqual(type(arr), list)
        self.assertEqual(len(arr), 2)
        arr = arr + fn2
        self.assertEqual(type(arr), list)
        self.assertEqual(len(arr), 3)



from .prox_fn import ProxFn
import numpy as np

def sum_entries(lin_op, **kwargs):
    c = np.ones(lin_op.shape)
    if "c" in kwargs:
        c += kwargs["c"]
    return zero_prox(lin_op, c=c, **kwargs)

class zero_prox(ProxFn):
    """The function f(x) = 0.
    """
    def _prox(self, rho, v, *args, **kwargs):
        """x = v
        """
        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return 0



from __future__ import print_function

#Includes
import subprocess
import os
from ctypes import cdll
import re
import ctypes
import sys
import numpy as np

class Halide(object):

    def __init__(self, generator_source = [], func=[],
                builddir=[], recompile=False, target='host', 
                generator_name=[], generator_param=[], 
                external_source =[], external_libs=[], 
                compile_flags=[], cleansource=True, verbose=False):

        """ Compiles and runs a halide pipeline defined in a generator file ``filepath``
            If recompile is not enabled, first, the library is searched and then loaded. 
            Otherwise it is recompiled and a new library is defined.

        Example:

            #Only compiles the source --> generates conv.so file 
            Halide('conv.cpp')

            #Compiles the source and runs function generator conv --> result is in numpy array output
            Halide('conv.cpp').run(A,K,output) 

            or 

            Halide('conv.cpp').conv(A,K,output) --> calls run

            #Only runs the precompiled halide function --> result is in numpy array output
            Halide('conv.cpp').conv(A,K,output)

        """

        #Use script location/build as default build directory
        if not builddir:
            builddir = os.path.join( os.path.dirname(os.path.realpath(__file__)), 'build' )
        
        #Save arguments
        self.generator_source = find_source(generator_source) #Try to find source file
        self.func = func
        self.builddir = builddir
        self.recompile = recompile
        self.target = target
        self.generator_name = generator_name
        self.generator_param = generator_param
        self.external_source = external_source
        self.external_libs = external_libs
        self.compile_flags = compile_flags
        self.cleansource = cleansource
        self.verbose = verbose

        #Output names that our halide compilation produces
        function_name, function_name_c, output_lib = output_names(self.func, self.generator_source, self.builddir)

        #Recompile if necessary
        if not os.path.exists(output_lib) or self.recompile:
            self.compile()

        #Syntactic sugar --> create method with name function_name
        setattr(self, function_name, lambda *args: self.run(*args) )

    def compile(self):

        #Generate the new code (exits on fail)
        gengen( self.generator_source, self.builddir, 
                self.target, self.generator_name, self.func, self.generator_param, 
                self.external_source, self.external_libs, self.compile_flags,
                self.cleansource, self.verbose )
            
    def run(self, *args):
        """ Execute Halide code that was compiled before. """

        #Output names that our halide compilation produces
        function_name, function_name_c, output_lib = output_names(self.func, self.generator_source, self.builddir)

        #Run
        args_ctype = convert_to_ctypes(args, function_name_c)

        # Directly load the lib
        launch = cdll.LoadLibrary(output_lib)

        # Call
        error = getattr(launch, function_name_c)(*args_ctype)

        if error != 0:
            print('Halide call to {0} returned {1}'.format(function_name_c, error), file=sys.stderr)
            exit()

class Params:
    """ Supported Params. """
    ImageParam_Float32 = ('ImageParam_Float32', np.float32)
    Param_Float32 = ('Param_Float32', np.float32)
    Param_Int32 = ('Param_Int32', np.int32)

def output_names(function_name, generator_source, builddir):

    #Define function name
    if not function_name:
        function_name = os.path.splitext(os.path.basename(generator_source))[0]

    if not function_name:
        print('Could not determine function name.', file=sys.stderr)
        exit()

    #Function name in c-launcher
    function_name_c = '{0}_c'.format(function_name)

    #Output library
    lib_name = os.path.join(builddir, function_name + '.so')

    return function_name, function_name_c, lib_name

def find_source( source ):

    source_found = source
    if not os.path.exists(source_found):

        #Try with path substitution
        sourcedir = os.path.dirname(source)
        sourcebase = os.path.basename(source)
        halide_dir = os.path.dirname(os.path.realpath(__file__))
        source_found = os.path.join(halide_dir, 'src', sourcebase)

        if not os.path.exists(source_found):

            #Try without path substitution
            halide_dir = os.path.dirname(os.path.realpath(__file__))
            source_found = os.path.join(halide_dir, 'src', source)

            if not os.path.exists(source_found):
                print('Error: Could not find source {0} in {1} or in {2}'.format(sourcebase,  sourcedir, halide_dir), file=sys.stderr)
                exit()  

    return source_found

def gengen( generator_source, builddir='./build',
            target='host', generator_name=[], function_name=[], generator_param=[], 
            external_source =[], external_libs=[], compile_flags = [],
            cleansource=True, verbose=True):
    
    """ Will take .cpp containing one (or more) Generators, compile them, link them with libHalide, and run
        the resulting executable to produce a .o/.h expressing the Generator's
        function. Function name is the C function name for the result """

    #Build directory
    if not os.path.exists(builddir):
        os.makedirs(builddir)

    #Generator code is a temporary file
    generator = os.path.join(builddir, 'gengen.XXXX')

    #File definitions
    halide_lib = '${HALIDE_PATH}/bin/libHalide.so'
    halid_incl = '-I${HALIDE_PATH}/include'
    generator_main = '${HALIDE_PATH}/tools/GenGen.cpp'

    #Define output names
    function_name, function_name_c, output_lib = output_names(function_name, generator_source, builddir)

    # It's OK for GENERATOR_NAME and FUNCTION_NAME to be empty
    # if the source we're compiling has only one generator registered,
    # we just use that one (and assume that FUNCTION_NAME=GENERATOR_NAME)
    generator_flag = ""
    if generator_name:
        generator_flag = "-g " + generator_name
        
    #Function flag
    function_flag = "-f " + function_name

    #Target flags
    target_flags = "target=" + target
    
    launcher_file = ''

    try:

        #Additional flags
        compile_flag_str = ''
        if compile_flags:
            for cf in compile_flags:
                compile_flag_str +=  cf + ' '

        #Compile
        cmd = 'g++ {0} -g -Wwrite-strings -std=c++11 -fno-rtti {1} {2} {3} {4} -lz -lpthread -ldl -o {5}'.format(
            compile_flag_str, halid_incl, generator_source, generator_main, halide_lib, generator)

        if verbose:
            print('Compiling {0}'.format(generator_source))
            print('\t' + cmd)
        subprocess.call(cmd, shell=True)

        #Run generator
        cmd = '{0} {1} {2} -o {3} {4}'.format(generator, generator_flag, function_flag, builddir, target_flags)
        if verbose:
            print('Calling generator')
            print('\t' + cmd)
        subprocess.call(cmd, shell=True)

        #Find params in output generated by generator
        header_file = os.path.join(builddir, function_name + '.h')
        object_file = os.path.join(builddir, function_name + '.o')
        params = scan_params(header_file, function_name, verbose)
        if verbose:
            print('Found {0} buffers and {1} float params and {2} int params'.format( 
                params.count(Params.ImageParam_Float32), params.count(Params.Param_Float32), params.count(Params.Param_Int32) ))

        #Generate launcher cpp and write
        launcher_file = os.path.join(builddir, function_name + '.cpp')
        launcher_body, argument_names = generate_launcher(header_file, function_name, function_name_c, params)
        with open(launcher_file, 'w') as fp:
            fp.write(launcher_body)

        #Compile launcher into library file (which will be called later by ctypes)
        if os.path.exists(output_lib):
            os.remove(output_lib)

        #External sources
        external_source_str= ''
        if external_source:
            for sc in external_source:
                external_source_str +=  find_source(sc) + ' '

        external_libs_str= ''
        if external_libs:
            for el in external_libs:
                external_libs_str +=  el + ' '
        
        cmd = 'g++ -fPIC -std=c++11 -Wall -O2 {0} {1} {2} -lpthread {3} -shared -o {4}'.format(launcher_file, 
                                                                    external_source_str, external_libs_str, object_file, output_lib)
        if verbose:
            print('Compiling library')
            print('\t' + cmd)
        subprocess.call(cmd, shell=True)

        return output_lib, function_name_c, argument_names

    except Exception as e:
        print('Error genererator compilation: {0}'.format(e.message), file=sys.stderr)
        exit()

    finally:
        # Cleanup
        if cleansource:
            source = [generator, header_file, object_file, launcher_file]
            for selem in source:
                if os.path.exists(selem):
                    os.remove(selem)

def convert_to_ctypes(args, func):
    """ Converts an argument list to a ctype compatible list for our launcher function """

    # pass numpy buffers using ctypes
    cargs = [];
    try:

        #Iterate over the args and convert
        for arg in args:

            #Check for supported types
            if isinstance(arg, np.ndarray):

                #Check for the array type
                if not arg.dtype == np.float32:
                    raise ValueError('Input array of type {0} detected, Not supported.'.format(arg.dtype) )

                #Otherwise add the bounds
                if len(arg.shape) > 4:
                    raise ValueError('Detected {0} dimensions. Halide supports only up to 4.'.format(len(arg.shape)) )

                #Check if fortran array 
                if len(arg.shape) > 1 and not np.isfortran(arg):
                    print('Arg ', arg)
                    raise ValueError('Currently supports only Fortran order' ) #Much faster and more natural halide code
                
                #Add ctype
                cargs.append(arg.ctypes.data_as(ctypes.c_void_p))

                #Add bound w,h,x,y ...
                for s in [1,0,2,3]:
                    if s < len(arg.shape):
                        cargs.append( ctypes.c_int( np.int32(arg.shape[s]) ))
                    else:
                        cargs.append( ctypes.c_int( np.int32(1) ))

            elif isinstance(arg, float) or isinstance(arg, np.float32):
                cargs.append( ctypes.c_float(arg) )

            elif isinstance(arg, int) or isinstance(arg, np.int32):
                cargs.append( ctypes.c_int(arg) )

            else:
                raise ValueError('Unsupported type.')

    except Exception as e:
        print('Error argument conversion: {0} in func {1}'.format(e.message, func), file=sys.stderr)
        exit()

    return cargs


def scan_params(header_file, function_name, verbose=True):
    """ Scans a header file from generator and reads params from compiled fucntion
    """

    #Find signature
    signature_start = 'int {}('.format(function_name)
    signature = ""
    searchfile = open(header_file, "r")
    for line in searchfile:
        if signature_start in line: 
            signature = line
    searchfile.close()

    #Verbose
    if verbose:
        print('Signature found: {}'.format(signature))

    #Find the argumetn list from the signature
    #Assumes halide checks internally if the arguments are correct (e.g. buffer_t with wrong elem size)
    arglist = []
    for arg_string in signature.split(","):
        if 'buffer_t *' in arg_string: 
            arglist.append( Params.ImageParam_Float32 )
        elif 'const float' in arg_string:
            arglist.append( Params.Param_Float32 )
        elif 'const int' in arg_string:
            arglist.append( Params.Param_Int32 )
        else:
            print('Error: Unsupported argument in {0}'.format(arg_string), file=sys.stderr)

    #Check for buffer count
    if arglist.count(Params.ImageParam_Float32) == 0:
        print('Error: No input buffers in halide generated header {0}'.format(header_file), file=sys.stderr)

    return arglist

def generate_launcher_arguments(params):

    #Generates buffer definitions for the launcher
    #Example
    #buffer_t b_input = {0, (uint8_t*)input, {width,height}, {1,width}, {0}, sizeof(float),};

    #The parameters are just numbered
    buffer_defs = [];
    argument_defs = [];
    argument_names = [];
    call_names = [];
    for id, param in enumerate(params):

        #Name of current argument
        argname = 'arg{:d}'.format(id)
        argument_names.append(argname)
        
        #Create buffer if buffer argument
        if param == Params.ImageParam_Float32:
            buffer_name = 'buf{0}'.format(argname)
            bbody = 'buffer_t {0}'.format(buffer_name)
            bbody += ' = {{0, (uint8_t*){0}, '.format(argname) #Float pointer

            #Linear strides
            #bbody += '{{ {0}w,{0}h,{0}s,{0}t }}, {{ 1, {0}w, {0}w * {0}h, {0}w * {0}h * {0}s }}, '.format(argname) #Extent and stride

            #Flip first two axis for natural buffer access (numpy like)
            bbody += '{{ {0}w,{0}h,{0}s,{0}t }}, {{ {0}h, 1, {0}w * {0}h, {0}w * {0}h * {0}s }}, '.format(argname) #Extent and stride

            bbody += '{0}, sizeof(float),};'

            #Add to buffer definitions
            buffer_defs.append(bbody)

            #Add argument and bounds
            argument_defs.append( 'float* {0}, int {0}w, int {0}h, int {0}s, int {0}t'.format(argname) )
            call_names.append( '&{0}'.format(buffer_name) )

        elif param == Params.Param_Float32:
            
            #Add argument
            argument_defs.append( 'const float {0}'.format(argname) )
            call_names.append( argname )

        elif param == Params.Param_Int32:
            #Add argument
            argument_defs.append( 'const int {0}'.format(argname) )
            call_names.append( argname )

    return argument_names, argument_defs, buffer_defs, call_names

def generate_launcher(header_file, function_name, function_name_c, params):

    """ Generates launcher glue code that runs generator from c-interface. 
        Similar to matlab runtime in halide. """

    #Header
    body = '// Launcher code for generator of function: {0}\n'.format(function_name)
    body += """#include <stdio.h>\n#include <stdlib.h>\n#include "{0}.h"\n\n#define LIBRARY_API extern "C"\n\n""".format(function_name)

    #Generate arguments
    argument_names, argument_defs, buffer_defs, call_names = generate_launcher_arguments(params);

    #Method signature
    function_def = 'LIBRARY_API int {0}('.format(function_name_c)
    body += function_def  #Contains at least one buffer
    for aid, argdef in enumerate(argument_defs):
        if aid == 0:
            body += argdef
        else:
            body += '\n' + (' ' * len(function_def)) + argdef
        body += ',' if aid < len(argument_defs) - 1 else '' #Trailing comma

    #Function body
    body += ')\n{\n'
    INDENT = '    '

    ## Add a buffer_t for each buffer
    body += INDENT + '//Transfer data from arrays into buffers\n'
    for buf in buffer_defs:
        body += INDENT + buf + '\n'
    body += '\n'

    #Arguments for the call
    call_argument = ''
    for cid, cn in enumerate(call_names):
        call_argument += cn;
        call_argument += ', ' if cid < len(call_names) - 1 else '' #Trailing comma


    #Call the function
    body += INDENT + '//Call our halide function with the buffers defined\n'
    body += INDENT + '{0}( {1} );\n\n'.format(function_name, call_argument )

    #Return
    body += INDENT + '// DONE\n'
    body += INDENT + 'return 0;\n'
    body += '}\n'

    return body, argument_names



# Utilities for getting the inverse of lin ops.
import numpy as np
from proximal.prox_fns import least_squares, sum_squares
from proximal.lin_ops import Variable, vstack
from proximal.utils import Impl

def get_dims(lin_ops):

    dims = -1
    for op in lin_ops:
        # Recurse on lin op tree.
        while True:
            if hasattr(op, 'dims'):
                if dims == -1:
                    dims = op.dims
                elif dims != op.dims:
                    raise Exception('Dims not consistent')
                    return -1

            if len(op.input_nodes) > 0:
                op = op.input_nodes[0]
            else:
                break

    if dims == -1:
        dims = None

    return dims

def get_implem(lin_ops):

    implem = Impl['numpy']
    for op in lin_ops:
        # Recurse on lin op tree.
        while True:
            if op.implementation != Impl['numpy']:
                implem = op.implementation
                break;

            if len(op.input_nodes) > 0:
                op = op.input_nodes[0]
            else:
                break

    return implem

def get_diag_quads(prox_fns, freq):
    """Returns all the quadratic functions that are Gram (freq) diagonal.
    """
    quad_funcs = [fn for fn in prox_fns if isinstance(fn, sum_squares)]
    if freq:
        return [fn for fn in quad_funcs if fn.lin_op.is_diag(freq=True) and \
                type(fn) == sum_squares]
    else:
        return [fn for fn in quad_funcs if fn.lin_op.is_diag(freq=False)]

def max_diag_set(prox_fns):
    """Return a maximal cardinality set of quadratic functions with
       diagonalizable lin ops.
    """
    freq_diag = get_diag_quads(prox_fns, True)
    spatial_diag = get_diag_quads(prox_fns, False)
    if len(spatial_diag) >= len(freq_diag):
        return spatial_diag
    else:
        return freq_diag

def get_least_squares_inverse(op_list, b, try_freq_diagonalize=True, verbose=False):
    if len(op_list) == 0:
        return None
    # Are all the operators diagonal?
    stacked = vstack(op_list)
    if stacked.is_gram_diag(freq=False):
        if verbose:
            print 'Optimized for diagonal inverse'

        diag = stacked.get_diag(freq=False).values()[0]
        diag = diag*np.conj(diag)
        x_update = least_squares(stacked, b, diag=diag)

    # Are all the operators diagonal in the frequency domain?
    elif try_freq_diagonalize and stacked.is_gram_diag(freq=True):

        diag = stacked.get_diag(freq=True).values()[0]
        diag = diag*np.conj(diag)
        dims = get_dims(op_list)
        implem = get_implem(op_list) #If any freqdiag is halide, solve with halide

        if verbose:
            dimstr = (' with dimensionality %d' % dims) if dims != None else ''
            print 'Optimized for diagonal frequency inverse' + dimstr

        x_update = least_squares(stacked, b,
                                 freq_diag=diag, freq_dims=dims, implem = implem)
    else:

        x_update = least_squares(stacked, b)

    return x_update



import abc
import numpy as np
from proximal.utils import Impl


def cast_to_const(expr):
    """Converts a non-LinOp to a Constant.
    """
    from .constant import Constant
    return expr if isinstance(expr, LinOp) else Constant(expr)


class LinOp(object):
    """Represents a linear operator.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, input_nodes, shape, implem=None):
        self.input_nodes = [cast_to_const(node) for node in input_nodes]
        self.shape = self.format_shape(shape)
        if implem is not None:
            self.set_implementation(implem)
        else:
            self.implementation = Impl['numpy']
        super(LinOp, self).__init__()

    def format_shape(self, shape):
        """Cast shape to a tuple.
        """
        # Convert scalars to tuples.
        if np.isscalar(shape):
            shape = (shape,)
        return shape

    def variables(self):
        """Return the list of variables used in the LinOp.
        """
        vars_ = []
        for arg in self.input_nodes:
            vars_ += arg.variables()
        unordered = list(set(vars_))  # Make unique, order by uuid.
        return sorted(unordered, key=lambda x: x.uuid)

    def constants(self):
        """Returns a list of constants in the LinOp.
        """
        consts = []
        for arg in self.input_nodes:
            consts += arg.constants()
        return consts

    def is_constant(self):
        """Is the LinOp constant?
        """
        return len(self.variables()) == 0

    def set_implementation(self, im):
        if im in Impl.values():
            self.implementation = im
        elif im in Impl.keys():
            self.implementation = Impl[im]
        else:
            raise Exception("Invalid implementation.")

        return self.implementation

    def implementation(self):
        return self.implementation

    @property
    def size(self):
        return np.prod(self.shape)

    @abc.abstractmethod
    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        return NotImplemented

    @abc.abstractmethod
    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        return NotImplemented

    def is_gram_diag(self, freq=False):
        """Is the lin op's Gram matrix diagonal (in the frequency domain)?
        """
        return self.is_diag(freq)

    def is_diag(self, freq=False):
        """Is the lin op diagonal (in the frequency domain)?
        """
        return False

    def get_diag(self, freq=False):
        """Returns the diagonal representation (A^TA)^(1/2).

        Parameters
        ----------
        freq : bool
            Is the diagonal representation in the frequency domain?
        Returns
        -------
        dict of variable to ndarray
            The diagonal operator acting on each variable.
        """
        return NotImplemented

    def __add__(self, other):
        """Lin Op + Lin Op.
        """
        other = cast_to_const(other)
        from .sum import sum
        args = []
        for elem in [self, other]:
            if isinstance(elem, sum):
                args += elem.input_nodes
            else:
                args += [elem]
        return sum(args)

    def __mul__(self, other):
        """Lin Op * Number.
        """
        from .scale import scale
        # Can only divide by scalar constants.
        if np.isscalar(other):
            return scale(other, self)
        else:
            raise TypeError("Can only multiply by a scalar constant.")

    def __rmul__(self, other):
        """Called for Number * Lin Op.
        """
        return self * other

    def __truediv__(self, other):
        """Lin Op / Number.
        """
        return self.__div__(other)

    def __div__(self, other):
        """Lin Op / Number.
        """
        from .scale import scale
        # Can only divide by scalar constants.
        if np.isscalar(other):
            return scale(1./other, self)
        else:
            raise TypeError("Can only divide by a scalar constant.")

    def __sub__(self, other):
        """Called for lin op - other.
        """
        return self + -other

    def __rsub__(self, other):
        """Called for other - lin_op.
        """
        return -self + other

    def __neg__(self):
        """The negation of the Lin Op.
        """
        return -1*self

    def __str__(self):
        """Default to string is name of class.
        """
        return self.__class__.__name__

    @property
    def value(self):
        inputs = []
        for node in self.input_nodes:
            inputs.append(node.value)
        output = np.zeros(self.shape)
        self.forward(inputs, [output])
        return output

    def get_offset(self):
        """Get the constant offset.
        """
        old_vals = {}
        for var in self.variables():
            old_vals[var] = var.value
            var.value = np.zeros(var.shape)
        offset = self.value
        # Restore old variable values.
        for var in self.variables():
            var.value = old_vals[var]
        return offset

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return NotImplemented



from __future__ import print_function, division

#Imports
from PIL import Image
import numpy as np
from numpy.fft import fftn, ifftn, fft2, ifft2
import cv2
import timeit
import sys

###############################################################################
## Implementations supported
###############################################################################

Impl = {'numpy': 0, 'halide': 1}

###############################################################################
## TODO: DIRTY HACK FOR BACKWARDS COMPATIBILITY!
###############################################################################

try:
    np.stack(np.array([1]))
except:
    from numpy.core import numeric
    def _stack(arrays, axis=0):
        arrays = [np.asanyarray(arr) for arr in arrays]
        if not arrays:
            raise ValueError('need at least one array to stack')

        shapes = set(arr.shape for arr in arrays)
        if len(shapes) != 1:
            raise ValueError('all input arrays must have the same shape')

        result_ndim = arrays[0].ndim + 1
        if not -result_ndim <= axis < result_ndim:
            msg = 'axis {0} out of bounds [-{1}, {1})'.format(axis, result_ndim)
            raise np.IndexError(msg)
        if axis < 0:
            axis += result_ndim

        sl = (slice(None),) * axis + (numeric.newaxis,)
        expanded_arrays = [arr[sl] for arr in arrays]
        return numeric.concatenate(expanded_arrays, axis=axis)
    np.stack = _stack

###############################################################################
## Image utils
###############################################################################

def im2nparray(img, datatype=np.float32):

    """ Converts and normalizes image in certain datatype (e.g. np.float32) """
    np_img = np.array(img)

    i = np.iinfo(np_img.dtype)
    max_class = 2 ** i.bits - 1
    np_img = np_img.astype(datatype) / max_class

    return np_img

###############################################################################
## Timing utils
###############################################################################

#Store last time stamp globally
global lastticstamp
lastticstamp = []

def tic():
    """ Default timer
    Example: t = tic()
         ... code
         elapsed = toc(t)
         print( '{0}: {1:.4f}ms'.format(message, elapsed) )
    """
    global lastticstamp

    t = timeit.default_timer()
    lastticstamp = t
    return t;

def toc(t = []):
    """ See tic f
    """

    global lastticstamp

    #Last tic
    if not t:
      if lastticstamp:
        t = lastticstamp
      else:
        print('Error: Call to toc did never call tic before.', file=sys.stderr)
        return 0.0

    #Measure time in ms
    elapsed = (timeit.default_timer() - t) * 1000.0  #in ms
    return elapsed

###############################################################################
## FFT utils
###############################################################################

def fftd(I, dims=None):

    #Compute fft
    if dims is None:
        X = fftn(I)
    elif dims == 2:
        X = fft2(I, axes=(0,1))
    else:
        X = fftn(I, axes=tuple(range(dims)) )

    return X

def ifftd(I, dims=None):

    #Compute fft
    if dims is None:
        X = ifftn(I)
    elif dims == 2:
        X = ifft2(I, axes=(0,1))
    else:
        X = ifftn(I, axes=tuple(range(dims)) )

    return X

def circshift(x, shifts):

    for j in range(len(shifts)):
        x = np.roll(x, shifts[j], axis=j)

    return x

def psf2otf(K, outsize, dims=None):

    #Size
    sK = K.shape
    assert len(sK) == len(outsize)

    #Pad to large size and circshift
    padfull = []
    for j in range(len(sK)):
        padfull.append( (0, outsize[j] - sK[j]) )

    Kfull = np.pad(K, padfull, mode='constant', constant_values=0.0)

    #Circular shift
    shifts = -np.floor_divide( np.array(sK), 2 )
    if dims is not None and dims < len(sK):
        shifts = shifts[0:dims]

    Kfull = circshift(Kfull, shifts )

    #Compute otf
    otf = fftd(Kfull, dims)

    # Estimate the rough number of operations involved in the computation of the FFT.
    if dims is not None and dims < len(sK):
        sK = sK[0:dims]

    nElem = np.prod( sK )
    nOps  = 0;
    for k in range(len(sK)):
        nffts = nElem / sK[k];
        nOps  = nOps + sK[k] * np.log2( sK[k] ) * nffts;

    # Discard the imaginary part of the psf if it's withi roundoff error.
    eps = np.finfo(np.float32).eps
    if np.amax( np.absolute( otf.imag ) ) / np.amax( np.absolute( otf ) ) <= nOps * eps :
        otf = otf.real

    return otf

###############################################################################
## Image metrics
###############################################################################

def psnr(x, ref, pad=None, maxval = 1.0):

    #Sheck size
    if ref.shape != x.shape:
        raise Exception("Wrong size in PSNR evaluation.")

    #Remove padding if necessary
    if pad is not None:

        ss = x.shape
        il = ()
        for j in range(len(ss)):
            if len(pad) >= j + 1 and pad[j] > 0:
                currpad = pad[j] 
                il += np.index_exp[ currpad:-currpad ]
            else:
                il += np.index_exp[:]

        mse = np.mean( (x[il] - ref[il])**2 )
    else:
        mse = np.mean( (x - ref)**2 )

    #MSE
    if mse > np.finfo(float).eps:
        return 10.0 * np.log10( maxval**2 / mse )
    else:
        return np.inf

###############################################################################
## Noise estimation
###############################################################################

#Currently only implements one method
NoiseEstMethod = {'daub_reflect': 0, 'daub_replicate': 1}

def estimate_std(z, method = 'daub_reflect'):
    #Estimates noise standard deviation assuming additive gaussian noise

    #Check method
    if (method not in NoiseEstMethod.values()) and (method in NoiseEstMethod.keys()):
        method = NoiseEstMethod[method]
    else:
        raise Exception("Invalid noise estimation method.")

    #Check shape
    if len(z.shape) == 2:
        z = z[..., np.newaxis]
    elif len(z.shape) != 3:
        raise Exception("Supports only up to 3D images.")

    #Run on multichannel image
    channels = z.shape[2]
    dev = np.zeros( channels )

    #Iterate over channels
    for ch in range(channels):

        #Daubechies denoising method
        if method == NoiseEstMethod['daub_reflect'] or method == NoiseEstMethod['daub_replicate']:
            daub6kern= np.array([0.03522629188571, 0.08544127388203, -0.13501102001025, \
                                -0.45987750211849, 0.80689150931109, -0.33267055295008], \
                                dtype=np.float32, order='F')

            if method == NoiseEstMethod['daub_reflect']:
                wav_det = cv2.sepFilter2D(z, -1, daub6kern, daub6kern, borderType=cv2.BORDER_REFLECT_101)
            else: 
                wav_det = cv2.sepFilter2D(z, -1, daub6kern, daub6kern, borderType=cv2.BORDER_REPLICATE)

            dev[ch] = np.median( np.absolute(wav_det) ) / 0.6745
        
    #Return standard deviation
    return dev



from __future__ import division
from proximal.lin_ops import CompGraph, scale, Variable, vstack
from proximal.prox_fns import least_squares
from proximal.utils.timings_log import TimingsLog, TimingsEntry
from invert import get_least_squares_inverse, get_diag_quads
from absorb import absorb_offset
import numpy as np

def partition(prox_fns, try_diagonalize=True):
    """Divide the proxable functions into sets Psi and Omega.
    """
    # Merge these quadratic functions with the v update.
    quad_funcs = []
    # All lin ops must be gram diagonal for the least squares problem
    # to be diagonal.
    func_opts = {True:[], False:[]}
    for freq in [True, False]:
        if all([fn.lin_op.is_gram_diag(freq) for fn in prox_fns]):
            func_opts[freq] = get_diag_quads(prox_fns, freq)
    # Quad funcs is the max cardinality set.
    if len(func_opts[True]) >= len(func_opts[False]):
        quad_funcs = func_opts[True]
    else:
        quad_funcs = func_opts[False]
    psi_fns = [fn for fn in prox_fns if fn not in quad_funcs]
    return psi_fns, quad_funcs

def solve(psi_fns, omega_fns, rho=1.0,
         max_iters=1000, eps_abs=1e-3, eps_rel=1e-3, x0 = None,
         lin_solver="cg", lin_solver_options=None,
         try_diagonalize=True, try_fast_norm=False,
         scaled=True,
         metric=None, convlog=None, verbose=0):
    prox_fns = psi_fns + omega_fns
    stacked_ops = vstack([fn.lin_op for fn in psi_fns])
    K = CompGraph(stacked_ops)
    # Rescale so (rho/2)||x - b||^2_2
    rescaling = np.sqrt(2./rho)
    quad_ops = []
    const_terms = []
    for fn in omega_fns:
        fn = fn.absorb_params()
        quad_ops.append( scale(rescaling*fn.beta, fn.lin_op) )
        const_terms.append( fn.b.flatten()*rescaling )
    # Check for fast inverse.
    op_list = [fn.lin_op for fn in psi_fns] + quad_ops
    stacked_ops = vstack(op_list)

    # Get optimize inverse (tries spatial and frequency diagonalization)
    v_update = get_least_squares_inverse(op_list, None, try_diagonalize, verbose)

    # Initialize everything to zero.
    v = np.zeros(K.input_size)
    z = np.zeros(K.output_size)
    u = np.zeros(K.output_size)

    #Initialize
    if x0 is not None:
        v[:] = np.reshape( x0, K.input_size )
        K.forward(v, z);

    # Buffers.
    Kv = np.zeros(K.output_size)
    KTu = np.zeros(K.input_size)
    s = np.zeros(K.input_size)

    # Log for prox ops.
    prox_log = TimingsLog(prox_fns)
    # Time iterations.
    iter_timing = TimingsEntry("ADMM iteration")
    # Convergence log for initial iterate
    if convlog != None:
        K.update_vars(v)
        objval = sum([fn.value for fn in prox_fns])
        convlog.record_objective( objval )
        convlog.record_timing( 0.0 )

    for i in range(max_iters):
        iter_timing.tic()
        if convlog != None:
            convlog.tic()

        z_prev = z.copy()

        # Update v.
        tmp = np.hstack([z-u] + const_terms)
        v = v_update.solve(tmp, x_init=v, lin_solver=lin_solver, options=lin_solver_options)

        # Update z.
        K.forward(v, Kv)
        Kv_u = Kv + u
        offset = 0
        for fn in psi_fns:
            slc = slice(offset, offset+fn.lin_op.size, None)
            Kv_u_slc = np.reshape(Kv_u[slc], fn.lin_op.shape)
            # Apply and time prox.
            prox_log[fn].tic()
            z[slc] = fn.prox(rho, Kv_u_slc, i).flatten()
            prox_log[fn].toc()
            offset += fn.lin_op.size
        # Update u.
        u += Kv - z
        K.adjoint(u, KTu)

        # Check convergence.
        r = Kv - z
        K.adjoint(rho*(z - z_prev), s)
        eps_pri = np.sqrt(K.output_size)*eps_abs + eps_rel*max([np.linalg.norm(Kv), np.linalg.norm(z)])
        eps_dual = np.sqrt(K.input_size)*eps_abs + eps_rel*np.linalg.norm(KTu)/rho

        #Convergence log
        if convlog != None:
            convlog.toc()
            K.update_vars(v)
            objval = sum([fn.value for fn in prox_fns])
            convlog.record_objective( objval )

        #Show progess
        if verbose > 0:
            #Evaluate objective only if required (expensive !)
            objstr = ''
            if verbose == 2:
                K.update_vars(v)
                objstr = ", obj_val = %02.03e" % sum([fn.value for fn in prox_fns])

            #Evaluate metric potentially
            metstr = '' if metric is None else ", {}".format( metric.message(v) )
            print "iter %d: ||r||_2 = %.3f, eps_pri = %.3f, ||s||_2 = %.3f, eps_dual = %.3f%s%s" % (
                i, np.linalg.norm(r), eps_pri, np.linalg.norm(s), eps_dual, objstr, metstr)

        iter_timing.toc()
        # Exit if converged.
        if np.linalg.norm(r) <= eps_pri and np.linalg.norm(s) <= eps_dual:
            break

    # Print out timings info.
    if verbose > 0:
        print iter_timing
        print "prox funcs:"
        print prox_log
        print "K forward ops:"
        print K.forward_log
        print "K adjoint ops:"
        print K.adjoint_log

    # Assign values to variables.
    K.update_vars(v)
    # Return optimal value.
    return sum([fn.value for fn in prox_fns])



from .prox_fn import ProxFn
import numpy as np
from proximal.utils.utils import *
from proximal.halide.halide import *

class group_norm1(ProxFn):
    """
    The function || ||x||_g ||_1, with g being a set of dimensions.
    This isotropic group sparsity proximal operator allows for example
    for isotropic TV regularization as a simple example.
    """

    def __init__(self, lin_op, group_dims, **kwargs):
        """Initialize temporary variables for _prox method.
        """
        self.group_dims = group_dims
        self.v_group_norm = np.zeros(lin_op.shape, dtype=float)

        #Temp array for halide
        self.tmpout = None
        if len(lin_op.shape) in [3,4] and lin_op.shape[-1] == 2 and self.group_dims == [len(lin_op.shape)-1] :
            self.tmpout = np.zeros( (lin_op.shape[0], lin_op.shape[1],
                                     lin_op.shape[2] if (len(lin_op.shape) > 3) else 1, 2) ,
                                    dtype=np.float32, order='FORTRAN');

        super(group_norm1, self).__init__(lin_op, **kwargs)

    def _prox(self, rho, v, *args, **kwargs):
        """x = v *  (1 - (1/rho) * 1/||x||_g )_+
        """

        if self.implementation == Impl['halide'] and \
           len(self.lin_op.shape) in [3,4] and self.lin_op.shape[-1] == 2 and self.group_dims == [len(self.lin_op.shape)-1]:

            #Halide implementation
            if len(self.lin_op.shape) == 3:
                tmpin = np.asfortranarray( np.reshape(v, (self.lin_op.shape[0], self.lin_op.shape[1],1,2)).astype(np.float32) )
            else:
                tmpin = np.asfortranarray( v.astype(np.float32) )

            Halide('prox_IsoL1.cpp').prox_IsoL1( tmpin, 1.0/rho, self.tmpout ) #Call
            np.copyto(v, np.reshape( self.tmpout, self.lin_op.shape ) )

        else:

            #Numpy implementation
            np.multiply(v,v, self.v_group_norm )

            #Sum along dimensions and keep dimensions
            orig_s = v.shape
            for d in self.group_dims:
                self.v_group_norm = np.sum(self.v_group_norm, axis=d, keepdims=True)

            #Sqrt
            np.sqrt(self.v_group_norm, self.v_group_norm)

            #Replicate
            tiles = ()
            for d in range( len(orig_s) ):
                if d in self.group_dims:
                    tiles += (orig_s[d],)
                else:
                    tiles += (1,)

            self.v_group_norm = np.tile(self.v_group_norm, tiles)

            #Thresholded group norm
            with np.errstate(divide='ignore'):
                np.maximum( 0.0, 1.0 - (1.0/rho) * (1.0/self.v_group_norm), self.v_group_norm )

            #Mult
            v *= self.v_group_norm

        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """

        #Square
        vsum = v.copy();
        np.multiply(v,v, vsum )

        #Sum along dimensions and keep dimensions
        orig_s = v.shape
        for d in self.group_dims:
            vsum = np.sum(vsum, axis=d, keepdims=True)

        #Sqrt
        np.sqrt(vsum, vsum)

        #L1 norm is then sum of norms
        return np.sum(vsum)

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.group_dims]

class weighted_group_norm1(group_norm1):
    """The function ||W.*x||_1.
    """
    def __init__(self, lin_op, group_dims, weight, **kwargs):
        self.weight = weight
        super(weighted_group_norm1, self).__init__(lin_op, group_dims, **kwargs)

    def _prox(self, rho, v, it):
        """x = v *  (1 - (|W|/rho) * /||x||_g )_+
        """

        #Square
        np.multiply(v,v, self.v_group_norm )

        #Sum along dimensions and keep dimensions
        orig_s = v.shape
        for d in self.group_dims:
            self.v_group_norm = np.sum(self.v_group_norm, axis=d, keepdims=True)

        #Sqrt
        np.sqrt(self.v_group_norm, self.v_group_norm)

        #Replicate
        tiles = ()
        for d in range( len(orig_s) ):
            if d in self.group_dims:
                tiles += (orig_s[d],)
            else:
                tiles += (1,)

        self.v_group_norm = np.tile(self.v_group_norm, tiles)

        #Thresholded group norm
        with np.errstate(divide='ignore'):
            np.maximum( 0.0, 1.0 - (np.absolute(self.weight)/rho) * (1.0/self.v_group_norm), self.v_group_norm )

        #Mult
        idxs = self.weight == 0
        np.copyto(v, v_tmp)
        v *= self.v_group_norm
        v[idxs] = v_tmp[idxs]
        return v

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return super(weighted_group_norm1, self)._eval(self.weight*v)

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.group_dims, self.weight]



from .prox_fn import ProxFn
import numpy as np
from scipy.optimize import fmin_l_bfgs_b

class diff_fn(ProxFn):
    """A generic prox operator for differentiable functions using L-BFGS.
    """
    def __init__(self, lin_op, func, fprime, bounds=None, factr=1e7, **kwargs):
        """Initialization.

        Args:
            lin_op: A linear operator.
            func: A function call for evaluating the function.
            fprime: A function call for evaluating the derivative.
            bounds: A list of (lower bound, upper bound) on each entry.
        """
        self.func = func
        self.fprime = fprime
        self.bounds = bounds
        self.factr = factr
        super(diff_fn, self).__init__(lin_op, **kwargs)

    def _prox(self, rho, v, *args, **kwargs):
        """Use L-BFGS
        """
        # Derivative of augmented function.
        prox_func = lambda x: self.func(x) + (rho/2.0)*np.square(x.ravel() - v.ravel()).sum()
        prox_fprime = lambda x: self.fprime(x).ravel() + rho*(x.ravel() - v.ravel())
        x, f, d = fmin_l_bfgs_b(prox_func, v, prox_fprime,
            bounds=self.bounds, factr=self.factr)
        return np.reshape(x, self.lin_op.shape)

    def _eval(self, v):
        """Evaluate the function on v (ignoring parameters).
        """
        return self.func(v)

    def get_data(self):
        """Returns info needed to reconstruct the object besides the args.

        Returns
        -------
        list
        """
        return [self.func, self.fprime, self.bounds, self.factr]



from __future__ import division
import abc
import numpy as np
from .utils import psnr

class metric(object):
    """Represents an metric for measuring reconstruction quality
    """
    __metaclass__ = abc.ABCMeta
    def __init__(self, desc, unit, decimals=None):
        self.desc = desc
        self.unit = unit
        self.decimals = decimals
        super(metric, self).__init__()

    @abc.abstractmethod
    def _eval(self, v):
        """Evaluate the metric
        """
        return NotImplemented

    def eval(self, v):
        """Evaluate the metric
        """
        if self.decimals is None:
            return self._eval( v )
        else:
            return np.round( self._eval( v ), decimals = self.decimals)
      
    def message(self, v):
        """Evaluate the metric
        """
        mval = self.eval(v)
        message = "{0}: {1} {2}".format(self.desc, mval, self.unit)
        return message

class psnr_metric(metric):
    """PSNR metric
    """
    def __init__(self, ref, maxval = 1.0, pad = None, decimals = 2):
        self.pad = pad
        self.ref = ref
        self.maxval = maxval
        super(psnr_metric, self).__init__("PSNR", "dB", decimals)

    def _eval(self, v):
        """Evaluate PSNR metric
        """
        return psnr( np.reshape(v,self.ref.shape), 
                    self.ref, 
                    pad=self.pad, maxval=self.maxval)


from proximal.tests.base_test import BaseTest
from proximal.lin_ops import *
from proximal.prox_fns import *
from proximal.algorithms import *
import cvxpy as cvx
import numpy as np
import scipy as sp

class TestAlgs(BaseTest):

    def test_admm(self):
        """Test ADMM algorithm.
        """
        X = Variable((10,5))
        B = np.reshape(np.arange(50), (10,5))*1.
        prox_fns = [sum_squares(X, b=B)]
        sltn = admm.solve(prox_fns, [], 1.0, eps_abs=1e-4, eps_rel=1e-4)
        self.assertItemsAlmostEqual(X.value, B, places=2)
        self.assertAlmostEqual(sltn, 0)

        prox_fns = [norm1(X, b=B, beta=2)]
        sltn = admm.solve(prox_fns, [], 1.0)
        self.assertItemsAlmostEqual(X.value, B/2., places=2)
        self.assertAlmostEqual(sltn, 0)

        prox_fns = [norm1(X), sum_squares(X, b=B)]
        sltn = admm.solve(prox_fns, [], 1.0)

        cvx_X = cvx.Variable(10, 5)
        cost = cvx.sum_squares(cvx_X - B) + cvx.norm(cvx_X, 1)
        prob = cvx.Problem(cvx.Minimize(cost))
        prob.solve()
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value)

        psi_fns, omega_fns = admm.partition(prox_fns)
        sltn = admm.solve(psi_fns, omega_fns, 1.0)
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value)

        prox_fns = [norm1(X)]
        quad_funcs = [sum_squares(X, b=B)]
        sltn = admm.solve(prox_fns, quad_funcs, 1.0)
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value)

        # With parameters for sum_squares
        prox_fns = [norm1(X)]
        quad_funcs = [sum_squares(X, b=B, alpha=0.1, beta=2., gamma=1, c=B)]
        sltn = admm.solve(prox_fns, quad_funcs, 1.0)

        cvx_X = cvx.Variable(10, 5)
        cost = 0.1*cvx.sum_squares(2*cvx_X - B) + cvx.sum_squares(cvx_X) + \
               cvx.norm(cvx_X, 1) + cvx.trace(B.T*cvx_X)
        prob = cvx.Problem(cvx.Minimize(cost))
        prob.solve()
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value, places=3)

        prox_fns = [norm1(X)]
        quad_funcs = [sum_squares(X - B, alpha=0.1, beta=2., gamma=1, c=B)]
        quad_funcs[0] = absorb_offset(quad_funcs[0])
        sltn = admm.solve(prox_fns, quad_funcs, 1.0)
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value, places=3)

        prox_fns = [norm1(X)]
        cvx_X = cvx.Variable(10, 5)
        # With linear operators.
        kernel = np.array([1,2,3])
        x = Variable(3)
        b = np.array([-41,413,2])
        prox_fns = [nonneg(x), sum_squares(conv(kernel, x), b=b)]
        sltn = admm.solve(prox_fns, [], 1.0, eps_abs=1e-5, eps_rel=1e-5)

        kernel_mat = np.matrix("2 1 3; 3 2 1; 1 3 2")
        cvx_X = cvx.Variable(3)
        cost = cvx.norm(kernel_mat*cvx_X - b)
        prob = cvx.Problem(cvx.Minimize(cost), [cvx_X >= 0])
        prob.solve()
        self.assertItemsAlmostEqual(x.value, cvx_X.value, places=2)
        self.assertAlmostEqual(np.sqrt(sltn), prob.value, places=2)

        prox_fns = [nonneg(x)]
        quad_funcs = [sum_squares(conv(kernel, x), b=b)]
        sltn = admm.solve(prox_fns, quad_funcs, 1.0, eps_abs=1e-5, eps_rel=1e-5)
        self.assertItemsAlmostEqual(x.value, cvx_X.value, places=2)
        self.assertAlmostEqual(np.sqrt(sltn), prob.value, places=2)

    def test_pock_chambolle(self):
        """Test pock chambolle algorithm.
        """
        X = Variable((10,5))
        B = np.reshape(np.arange(50), (10,5))
        prox_fns = [sum_squares(X, b=B)]
        sltn = pc.solve(prox_fns, [], 1.0, 1.0, 1.0)
        self.assertItemsAlmostEqual(X.value, B, places=2)
        self.assertAlmostEqual(sltn, 0)

        prox_fns = [norm1(X, b=B, beta=2)]
        sltn = pc.solve(prox_fns, [], 1.0, 1.0, 1.0)
        self.assertItemsAlmostEqual(X.value, B/2., places=2)
        self.assertAlmostEqual(sltn, 0, places=2)

        prox_fns = [norm1(X), sum_squares(X, b=B)]
        sltn = pc.solve(prox_fns, [], 0.5, 1.0, 1.0)

        cvx_X = cvx.Variable(10, 5)
        cost = cvx.sum_squares(cvx_X - B) + cvx.norm(cvx_X, 1)
        prob = cvx.Problem(cvx.Minimize(cost))
        prob.solve()
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value)

        psi_fns, omega_fns = pc.partition(prox_fns)
        sltn = pc.solve(psi_fns, omega_fns, 0.5, 1.0, 1.0,
            eps_abs=1e-5, eps_rel=1e-5)
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value)

        # With linear operators.
        kernel = np.array([1,2,3])
        kernel_mat = np.matrix("2 1 3; 3 2 1; 1 3 2")
        L = np.linalg.norm(kernel_mat)
        x = Variable(3)
        b = np.array([-41,413,2])
        prox_fns = [nonneg(x), sum_squares(conv(kernel, x), b=b)]
        sltn = pc.solve(prox_fns, [], 0.1, 0.1, 1.0, max_iters=3000,
            eps_abs=1e-5, eps_rel=1e-5)
        cvx_X = cvx.Variable(3)
        cost = cvx.norm(kernel_mat*cvx_X - b)
        prob = cvx.Problem(cvx.Minimize(cost), [cvx_X >= 0])
        prob.solve()
        self.assertItemsAlmostEqual(x.value, cvx_X.value, places=2)

        psi_fns, omega_fns = pc.partition(prox_fns)
        sltn = pc.solve(psi_fns, omega_fns, 0.1, 0.1, 1.0, max_iters=3000,
            eps_abs=1e-5, eps_rel=1e-5)
        self.assertItemsAlmostEqual(x.value, cvx_X.value, places=2)

        # # TODO
        # # Multiple variables.
        # x = Variable(1)
        # y = Variable(1)
        # prox_fns = [nonneg(x), sum_squares(vstack([x,y]), b=np.arange(2))]
        # sltn = pc(prox_fns, [prox_fns[-1]], 0.1, 0.1, 1.0,
        #     max_iters=3000,  eps_abs=1e-5, eps_rel=1e-5, try_diagonalize=False)
        # self.assertItemsAlmostEqual(x.value, [0])
        # self.assertItemsAlmostEqual(y.value, [1])

        # sltn = pc(prox_fns, [prox_fns[-1]], 0.1, 0.1, 1.0,
        #     max_iters=3000,  eps_abs=1e-5, eps_rel=1e-5, try_diagonalize=True)
        # self.assertItemsAlmostEqual(x.value, [0])
        # self.assertItemsAlmostEqual(y.value, [1])

    def test_half_quadratic_splitting(self):


        """Test half quadratic splitting.
        """
        X = Variable((10,5))
        B = np.reshape(np.arange(50), (10,5))
        prox_fns = [sum_squares(X, b=B)]
        sltn = hqs.solve(prox_fns, [], eps_rel=1e-4, max_iters = 100, max_inner_iters = 50)
        self.assertItemsAlmostEqual(X.value, B, places=2)
        self.assertAlmostEqual(sltn, 0)

        prox_fns = [norm1(X, b=B, beta=2)]
        sltn = hqs.solve(prox_fns, [], eps_rel=1e-4, max_iters = 100, max_inner_iters = 50)
        self.assertItemsAlmostEqual(X.value, B/2., places=2)
        self.assertAlmostEqual(sltn, 0)

        prox_fns = [norm1(X), sum_squares(X, b=B)]
        sltn = hqs.solve(prox_fns, [], eps_rel=1e-7,
                     rho_0 = 1.0, rho_scale = np.sqrt(2.0) * 2.0, rho_max = 2**16,
                    max_iters = 20, max_inner_iters = 500)

        cvx_X = cvx.Variable(10, 5)
        cost = cvx.sum_squares(cvx_X - B) + cvx.norm(cvx_X, 1)
        prob = cvx.Problem(cvx.Minimize(cost))
        prob.solve()
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value, places=3)

        psi_fns, omega_fns = hqs.partition(prox_fns)
        sltn = hqs.solve(psi_fns, omega_fns, eps_rel=1e-7,
                     rho_0 = 1.0, rho_scale = np.sqrt(2.0) * 2.0, rho_max = 2**16,
                    max_iters = 20, max_inner_iters = 500)
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value, places=3)

        # With linear operators.
        kernel = np.array([1,2,3])
        kernel_mat = np.matrix("2 1 3; 3 2 1; 1 3 2")
        L = np.linalg.norm(kernel_mat)
        x = Variable(3)
        b = np.array([-41,413,2])
        prox_fns = [nonneg(x), sum_squares(conv(kernel, x), b=b)]
        hqs.solve(prox_fns, [], eps_rel=1e-9, rho_0 = 4, rho_scale = np.sqrt(2.0)*1.0,
                  rho_max = 2**16, max_iters = 30, max_inner_iters = 500)

        cvx_X = cvx.Variable(3)
        cost = cvx.norm(kernel_mat*cvx_X - b)
        prob = cvx.Problem(cvx.Minimize(cost), [cvx_X >= 0])
        prob.solve()
        self.assertItemsAlmostEqual(x.value, cvx_X.value, places=0)

        psi_fns, omega_fns = hqs.partition(prox_fns)
        hqs.solve(psi_fns, omega_fns, eps_rel=1e-9, rho_0 = 4, rho_scale = np.sqrt(2.0)*1.0,
                  rho_max = 2**16, max_iters = 30, max_inner_iters = 500)
        self.assertItemsAlmostEqual(x.value, cvx_X.value, places=0)

    def test_lin_admm(self):
        """Test linearized admm. algorithm.
        """
        X = Variable((10,5))
        B = np.reshape(np.arange(50), (10,5))
        prox_fns = [sum_squares(X, b=B)]
        sltn = ladmm.solve(prox_fns, [], 0.1, max_iters=500)
        self.assertItemsAlmostEqual(X.value, B, places=2)
        self.assertAlmostEqual(sltn, 0)

        prox_fns = [norm1(X, b=B, beta=2)]
        sltn = ladmm.solve(prox_fns, [], 0.1, max_iters=500)
        self.assertItemsAlmostEqual(X.value, B/2., places=2)
        self.assertAlmostEqual(sltn, 0, places=2)

        prox_fns = [norm1(X), sum_squares(X, b=B)]
        sltn = ladmm.solve(prox_fns, [], 0.1, max_iters=500)

        cvx_X = cvx.Variable(10, 5)
        cost = cvx.sum_squares(cvx_X - B) + cvx.norm(cvx_X, 1)
        prob = cvx.Problem(cvx.Minimize(cost))
        prob.solve()
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value)

        psi_fns, omega_fns = ladmm.partition(prox_fns)
        sltn = ladmm.solve(psi_fns, omega_fns, 0.1, max_iters=500)
        self.assertItemsAlmostEqual(X.value, cvx_X.value, places=2)
        self.assertAlmostEqual(sltn, prob.value)

        # With linear operators.
        kernel = np.array([1,2,3])
        kernel_mat = np.matrix("2 1 3; 3 2 1; 1 3 2")
        L = np.linalg.norm(kernel_mat)
        x = Variable(3)
        b = np.array([-41,413,2])
        prox_fns = [nonneg(x), sum_squares(conv(kernel, x), b=b)]
        sltn = ladmm.solve(prox_fns, [], 0.1, max_iters=3000, eps_abs=1e-5,
                           eps_rel=1e-5)

        cvx_X = cvx.Variable(3)
        cost = cvx.norm(kernel_mat*cvx_X - b)
        prob = cvx.Problem(cvx.Minimize(cost), [cvx_X >= 0])
        prob.solve()
        self.assertItemsAlmostEqual(x.value, cvx_X.value, places=2)

        psi_fns, omega_fns = ladmm.partition(prox_fns)
        sltn = ladmm.solve(psi_fns, omega_fns, 0.1, max_iters=3000, eps_abs=1e-5,
            eps_rel=1e-5)
        self.assertItemsAlmostEqual(x.value, cvx_X.value, places=2)

    def test_equil(self):
        """Test equilibration.
        """
        from proximal.algorithms.equil import newton_equil
        np.random.seed(1)
        kernel = np.array([1,1,1])/np.sqrt(3)
        kernel_mat = np.ones((3,3))/np.sqrt(3)
        x = Variable(3)
        wr = np.array([10, 5, 7])
        K = mul_elemwise(wr, x)
        K = conv(kernel, K)
        wl = np.array([100, 50, 3])
        K = mul_elemwise(wl, K)
        K = CompGraph(K)

        # Equilibrate
        gamma = 1e-1
        d, e = equil(K, 1000, gamma=gamma, M=5)
        tmp = d*wl*kernel_mat*wr*e
        u, v = np.log(d), np.log(e)
        obj_val = np.square(tmp).sum()/2 - u.sum() - v.sum() + \
        gamma*(np.linalg.norm(v)**2 + np.linalg.norm(u)**2)

        d, e = newton_equil(wl*kernel_mat*wr, gamma, 100)
        tmp = d*wl*kernel_mat*wr*e
        u, v = np.log(d), np.log(e)
        sltn_val = np.square(tmp).sum()/2 - u.sum() - v.sum() + \
        gamma*(np.linalg.norm(v)**2 + np.linalg.norm(u)**2)
        self.assertAlmostEqual((obj_val - sltn_val)/sltn_val, 0, places=3)



from __future__ import division
from proximal.lin_ops import CompGraph, est_CompGraph_norm, scale, Variable, vstack
from proximal.prox_fns import least_squares
from proximal.utils.timings_log import TimingsLog, TimingsEntry
from invert import get_least_squares_inverse, max_diag_set
import numpy as np
import warnings

def partition(prox_fns, try_diagonalize=True):
    """Divide the proxable functions into sets Psi and Omega.
    """
    # Omega must be a single function.
    # Merge quadratic functions into the x update.
    # Automatically try to split the problem.
    quad_fns = max_diag_set(prox_fns)
    split_fn = []
    omega_fns = []
    if len(quad_fns) == 0:
        for fn in prox_fns:
            if type(fn.lin_op) == Variable:
                split_fn = [fn]
                break
        omega_fns = split_fn
    else:
        # Proximal solve for:
        # G(x) + 1/(2*tau) * ||x - v||^2_2, with G containing all quadratics
        quad_ops = []
        const_terms = []
        for fn in quad_fns:
            fn = fn.absorb_params()
            quad_ops.append( fn.beta*fn.lin_op )
            const_terms.append( fn.b.flatten() )

        stacked_ops = vstack(quad_ops)
        b = np.hstack(const_terms)
        # Get optimize inverse (tries spatial and frequency diagonalization)
        x_update = get_least_squares_inverse(quad_ops, b, try_diagonalize)
        omega_fns = [x_update]

    psi_fns = [fn for fn in prox_fns if fn not in split_fn + quad_fns]
    return psi_fns, omega_fns

def solve(psi_fns, omega_fns, lmb=1.0, mu = None, quad_funcs=None,
          max_iters=1000, eps_abs=1e-3, eps_rel=1e-3,
          lin_solver="cg", lin_solver_options=None,
          try_diagonalize=True, try_fast_norm=True, scaled=False,
          metric=None, convlog=None, verbose=0):

    # Can only have one omega function.
    assert len(omega_fns) <= 1
    prox_fns = psi_fns + omega_fns
    stacked_ops = vstack([fn.lin_op for fn in psi_fns])
    K = CompGraph(stacked_ops)
    #Select optimal parameters if wanted
    if lmb is None or mu is None:
        lmb, mu = est_params_lin_admm(K, lmb, verbose, scaled, try_fast_norm)

    # Initialize everything to zero.
    v = np.zeros(K.input_size)
    z = np.zeros(K.output_size)
    u = np.zeros(K.output_size)

    # Buffers.
    Kv = np.zeros(K.output_size)
    KTu = np.zeros(K.input_size)
    s = np.zeros(K.input_size)

    Kvzu = np.zeros(K.output_size)
    vzu = np.zeros(K.input_size)
    v_prev = np.zeros(K.input_size)
    z_prev = np.zeros(K.output_size)

    # Log for prox ops.
    prox_log = TimingsLog(prox_fns)
    # Time iterations.
    iter_timing = TimingsEntry("LIN-ADMM iteration")
    #Convergence log for initial iterate
    if convlog != None:
        K.update_vars(v)
        objval = sum([fn.value for fn in prox_fns])
        convlog.record_objective( objval )
        convlog.record_timing( 0.0 )

    for i in range(max_iters):
        iter_timing.tic()
        if convlog != None:
            convlog.tic()

        v_prev[:] = v
        z_prev[:] = z

        # Update v
        K.forward(v, Kv)
        Kvzu[:] = Kv - z + u
        K.adjoint(Kvzu, v)
        v[:] = v_prev - (mu/lmb) * v

        if len(omega_fns) > 0:
            v[:] = omega_fns[0].prox(1.0/mu, v, x_init=v_prev.copy(),
                lin_solver=lin_solver, options=lin_solver_options)

        # Update z.
        K.forward(v, Kv)
        Kv_u = Kv + u
        offset = 0
        for fn in psi_fns:
            slc = slice(offset, offset+fn.lin_op.size, None)
            Kv_u_slc = np.reshape(Kv_u[slc], fn.lin_op.shape)
            # Apply and time prox.
            prox_log[fn].tic()
            z[slc] = fn.prox(1.0/lmb, Kv_u_slc, i).flatten()
            prox_log[fn].toc()
            offset += fn.lin_op.size

        # Update u.
        u += Kv - z
        K.adjoint(u, KTu)

        # Check convergence.
        r = Kv - z
        K.adjoint((1.0/lmb)*(z - z_prev), s)
        eps_pri = np.sqrt(K.output_size)*eps_abs + eps_rel*max([np.linalg.norm(Kv), np.linalg.norm(z)])
        eps_dual = np.sqrt(K.input_size)*eps_abs + eps_rel*np.linalg.norm(KTu)/(1.0/lmb)

        #Convergence log
        if convlog != None:
            convlog.toc()
            K.update_vars(v)
            objval = sum([fn.value for fn in prox_fns])
            convlog.record_objective( objval )

        #Show progess
        if verbose > 0:
            #Evaluate objective only if required (expensive !)
            objstr = ''
            if verbose == 2:
                K.update_vars(v)
                objstr = ", obj_val = %02.03e" % sum([fn.value for fn in prox_fns])

            #Evaluate metric potentially
            metstr = '' if metric is None else ", {}".format( metric.message(v) )
            print "iter %d: ||r||_2 = %.3f, eps_pri = %.3f, ||s||_2 = %.3f, eps_dual = %.3f%s%s" % (
                i, np.linalg.norm(r), eps_pri, np.linalg.norm(s), eps_dual, objstr, metstr)

        iter_timing.toc()
        if np.linalg.norm(r) <= eps_pri and np.linalg.norm(s) <= eps_dual:
            break

    # Print out timings info.
    if verbose > 0:
        print iter_timing
        print "prox funcs:"
        print prox_log
        print "K forward ops:"
        print K.forward_log
        print "K adjoint ops:"
        print K.adjoint_log

    # Assign values to variables.
    K.update_vars(v)

    # Return optimal value.
    return sum([fn.value for fn in prox_fns])

def est_params_lin_admm(K, lmb=None, verbose=True, scaled=False, try_fast_norm=False):

    #Select lambda
    lmb = 1.0 if lmb is None else np.maximum( lmb, 1e-5)

    #Warn user
    if lmb > 1.0:
        warnings.warn("Large lambda value given by user.")

    #Estimate Lipschitz constant and comput tau
    if scaled:
        L = 1
    else:
        L = est_CompGraph_norm(K, try_fast_norm)
    mu = lmb / (L**2)

    if verbose:
        print "Estimated params [lambda = %3.3f | mu = %3.3f | L_est = %3.4f]" % (lmb, mu, L)

    return lmb, mu



from .lin_op import LinOp
import numpy as np

class sum(LinOp):
    """Sums its inputs.
    """
    def __init__(self, input_nodes, implem=None):
        shape = input_nodes[0].shape
        super(sum, self).__init__(input_nodes, shape)

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        if len(inputs) > 1:
            np.copyto(outputs[0], np.sum(inputs, 0))
        else:
            np.copyto(outputs[0], inputs[0])

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        for output in outputs:
            np.copyto(output, inputs[0])

    def is_diag(self, freq=False):
        """Is the lin op diagonal (in the frequency domain)?
        """
        return all([arg.is_diag(freq) for arg in self.input_nodes])

    def get_diag(self, freq=False):
        """Returns the diagonal representation (A^TA)^(1/2).

        Parameters
        ----------
        freq : bool
            Is the diagonal representation in the frequency domain?
        Returns
        -------
        dict of variable to ndarray
            The diagonal operator acting on each variable.
        """
        var_diags = {var:np.zeros(var.size) for var in self.variables()}
        for arg in self.input_nodes:
            arg_diags = arg.get_diag(freq)
            for var, diag in arg_diags.items():
                var_diags[var] = var_diags[var]+diag
        return var_diags

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return np.sum(input_mags)

class copy(sum):
    def __init__(self, shape, implem=None):
        self.shape = shape
        self.input_nodes = []

    def forward(self, inputs, outputs):
        """The forward operator.

        Reads from inputs and writes to outputs.
        """
        super(copy, self).adjoint(inputs, outputs)

    def adjoint(self, inputs, outputs):
        """The adjoint operator.

        Reads from inputs and writes to outputs.
        """
        super(copy, self).forward(inputs, outputs)

    def norm_bound(self, input_mags):
        """Gives an upper bound on the magnitudes of the outputs given inputs.

        Parameters
        ----------
        input_mags : list
            List of magnitudes of inputs.

        Returns
        -------
        float
            Magnitude of outputs.
        """
        return input_mags[0]
